# SPDX-License-Identifier: MIT

from dataclasses import dataclass
import pprint
from enum import IntEnum

from ..common import *
from m1n1.utils import *
from construct import *
from m1n1.constructutils import Ver

@dataclass
class ByRef:
    val: object

class Pointer(Subconstruct):
    pass

class InPtr(Pointer):
    pass

class OutPtr(Pointer):
    pass

class InOutPtr(Pointer):
    pass

class InOut(Subconstruct):
    pass

Ptr = InOutPtr

class NULL:
    def __str__(self):
        return "NULL"
    def __repr__(self):
        return "NULL"
NULL = NULL()

class Method:
    def __init__(self, rtype, name, *args, **kwargs):
        self.rtype = rtype
        self.name = name

        if args and kwargs:
            raise Exception("Cannot specify args and kwargs")
        elif args:
            args = [(f"arg{i}", arg) for i, arg in enumerate(args)]
            self.as_kwargs = False
        elif kwargs:
            args = list(kwargs.items())
            self.as_kwargs = True
        else:
            args = []

        self.args = args

        in_size = 0
        out_size = 0
        self.in_fields = []
        self.out_fields = []

        if rtype is not None:
            args.append(("ret", rtype))

        self.dir = []
        self.nullable = []
        self.array_of_p = []

        for i, (name, field) in enumerate(self.args):
            align = 1

            pfield = field
            dir = "in"

            if name == "ret":
                dir = "out"

            while isinstance(pfield, Subconstruct):
                if isinstance(pfield, InPtr):
                    dir = "in"
                elif isinstance(pfield, OutPtr):
                    dir = "out"
                elif isinstance(pfield, (InOut, InOutPtr)):
                    dir = "inout"
                pfield = pfield.subcon
            if isinstance(pfield, FormatField):
                align = min(4, pfield.length)

            if dir in ("in", "inout"):
                #if in_size % align:
                    #self.in_fields.append(Padding(align - (in_size % align)))
                    #in_size += align - (in_size % align)

                self.in_fields.append(name / field)
                in_size += field.sizeof()

            if dir in ("out", "inout"):
                #if out_size % align:
                    #self.out_fields.append(Padding(align - (out_size % align)))
                    #out_size += align - (out_size % align)

                self.out_fields.append(name / field)
                out_size += field.sizeof()

            self.dir.append(dir)

        for i, (name, field) in enumerate(self.args):
            array_size = None
            array_of_p = False
            nullable = False
            pfield = field

            while isinstance(pfield, Subconstruct):
                if isinstance(pfield, Array) and array_size is None:
                    array_size = pfield.count
                if isinstance(pfield, Pointer):
                    nullable = True
                    array_of_p = array_size is not None
                pfield = pfield.subcon

            if nullable:
                if array_of_p:
                    self.in_fields.append((name + "_null") / bool_[array_size])
                    in_size += array_size
                else:
                    self.in_fields.append((name + "_null") / bool_)
                    in_size += 1

            self.nullable.append(nullable)
            self.array_of_p.append(array_of_p)

        if in_size % 4:
            self.in_fields.append(Padding(4 - (in_size % 4)))
        if out_size % 4:
            self.out_fields.append(Padding(4 - (out_size % 4)))

        self.in_struct = Struct(*self.in_fields)
        self.out_struct = Struct(*self.out_fields)

    def get_field_val(self, i, in_vals, out_vals=None, nullobj=None):
        name, field = self.args[i]

        nullable = self.nullable[i]
        array_of_p = self.array_of_p[i]

        val = None

        if out_vals:
            val = out_vals.get(name, val)
        if val is None and in_vals:
            val = in_vals.get(name, val)

        if nullable and val is not None:
            null = in_vals.get(name + "_null", None)
            if null is None:
                return None
            if not array_of_p:
                val = nullobj if null else val
            else:
                val2 = [nullobj if n else val for val, n in zip(val, null)]
                if isinstance(val, ListContainer):
                    val2 = ListContainer(val2)
                val = val2

        return val

    def fmt_args(self, in_vals, out_vals=None):
        s = []

        for i, (name, field) in enumerate(self.args):
            if name == "ret":
                continue

            dir = self.dir[i]
            nullable = self.nullable[i]

            val = self.get_field_val(i, in_vals, out_vals, nullobj=NULL)

            if val is not None:
                if self.is_long(val):
                    s.append(f"{name}=...")
                elif isinstance(val, ListContainer):
                    s.append(f"{name}={list(val)!r}")
                else:
                    s.append(f"{name}={val!r}")
            elif dir == "out":
                s.append(f"{name}=<out>")
            else:
                s.append(f"{name}=?")

        return ", ".join(s)

    def print_long_args(self, indent, in_vals, out_vals=None):
        for i, (name, field) in enumerate(self.args):
            if name == "ret":
                continue

            val = self.get_field_val(i, in_vals, out_vals, nullobj=NULL)

            if name in in_vals and out_vals is not None and name not in out_vals:
                continue

            if self.is_long(val):
                hdr = f"{indent}  {name} = "
                if isinstance(val, (ListContainer, Container)):
                    print(hdr + str(val).replace("\n", "\n" + indent))
                elif isinstance(val, bytes):
                    print(hdr + f"({len(val):#x} bytes)")
                    chexdump(val, indent=indent + "    ")
                else:
                    dindent = " " * len(hdr)
                    if isinstance(val, dict) and "_io" in val:
                        del val["_io"]
                    print(hdr + pprint.pformat(val, sort_dicts=False).replace("\n", "\n" + dindent))

    def is_long(self, arg):
        if isinstance(arg, (list, bytes)):
            return len(arg) > 4 or any(self.is_long(i) for i in arg)

        return isinstance(arg, (dict, list, bytes))

    def parse_input(self, data):
        vals = self.in_struct.parse(data)

        return Container({ k: v() if callable(v) else v for k,v in vals.items() })

    def parse_output(self, data, in_vals):
        context = dict(in_vals)

        if "data" in context:
            del context["data"]

        vals = self.out_struct.parse(data, **context)

        return Container({ k: v() if callable(v) else v for k,v in vals.items() })

    def __str__(self):
        if self.rtype is None:
            rtype = "void"
        else:
            rtype = str(self.rtype)

        args = []
        for name, field in self.args:
            if name == "ret":
                continue
            args.append(f"{field} {name}")

        return f"{rtype} {self.name}({', '.join(args)})"

    def callback(self, func, in_data):
        in_vals = self.parse_input(in_data)

        args = []
        kwargs = {}

        out_vals = {}

        for i, (name, field) in enumerate(self.args):
            if name == "ret":
                continue

            dir = self.dir[i]

            val = self.get_field_val(i, in_vals, out_vals, nullobj=NULL)
            is_null = val is NULL
            if is_null:
                val = None

            if dir == "inout":
                if val is not None and not isinstance(val, list):
                    val = ByRef(val)
                out_vals[name] = val
            elif dir == "out" and not is_null:
                val = ByRef(None)
                out_vals[name] = val

            if self.as_kwargs:
                kwargs[name] = val
            else:
                args.append(val)

        retval = func(*args, **kwargs)

        if self.rtype is None:
            assert retval is None
        else:
            assert retval is not None
            out_vals["ret"] = retval

        out_vals = {k: v.val if isinstance(v, ByRef) else v for k, v in out_vals.items()}

        context = dict(in_vals)

        if "obj" in context:
            del context["obj"]

        out_data = self.out_struct.build(out_vals, **context)
        return out_data


    def call(self, call, *args, **kwargs):
        if args and kwargs:
            raise Exception("Cannot use both args and kwargs")

        if args:
            for arg, (name, field) in zip(args, self.args):
                kwargs[name] = arg

        in_vals = {}
        out_refs = {}

        for i, (name, field) in enumerate(self.args):
            if name == "ret":
                continue

            val = kwargs[name]
            dir = self.dir[i]
            nullable = self.nullable[i]
            array_of_p = self.array_of_p[i]

            if nullable:
                if not array_of_p:
                    in_vals[name + "_null"] = val is None
                else:
                    defaults = field.parse(b"\x00" * field.sizeof())
                    in_vals[name + "_null"] = [i is None for i in val]
                    val = [v if v is not None else defaults[i] for i, v in enumerate(val)]
            else:
                assert val is not None

            if val is None:
                continue

            if dir == "out":
                assert isinstance(val, ByRef)
                out_refs[name] = val
            elif dir == "inout":
                if isinstance(val, ByRef):
                    in_vals[name] = val.val
                    out_refs[name] = val
                elif val is not None:
                    in_vals[name] = val
            elif val is not None:
                in_vals[name] = val

        in_data = self.in_struct.build(in_vals)
        print(f"{self.name}({self.fmt_args(in_vals)})")

        out_data = call(in_data)
        out_vals = self.parse_output(out_data, in_vals)

        for k, v in out_refs.items():
            v.val = out_vals[k]

        if self.rtype is not None:
            return out_vals["ret"]

def dump_fields(fields):
    off = 0
    for f in fields:
        sizeof = f.sizeof()
        print(f"{off:#x}: {f} ({sizeof:#x})")
        off += sizeof

class Call(Method):
    pass

class Callback(Method):
    pass

int8_t = Int8sl
uint8_t = Int8ul
int16_t = Int16sl
uint16_t = Int16ul
int32_t = Int32sl
uint32_t = Int32ul
int64_t = Int64sl
uint64_t = Int64ul

uint = uint32_t
int_ = int32_t
ulong = uint64_t
long_ = int64_t

void = None

class IPCObject:
    @classmethod
    def methods(cls):
        ret = {}
        for c in cls.mro():
            ret.update({k: (cls, v) for k, v in cls.__dict__.items() if isinstance(v, Method)})

        return ret

rt_bw_config_t = Struct(
    "unk1" / UnkBytes(8),
    "reg1" / Int64ul,
    "reg2" / Int64ul,
    "unk2" / UnkBytes(4),
    "bit" / Int32ul,
    "padding" / UnkBytes(0x1c),
)

frame_sync_props_t = Struct(
    "unk1" / UnkBytes(28),
)

IOUserClient = Struct(
    "addr" / Hex(Int64ul),
    "unk" / Int32ul,
    "flag1" / Int8ul,
    "flag2" / Int8ul,
    Padding(2)
)

IOMobileFramebufferUserClient = IOUserClient

IOMFBStatus = Int32ul
IOMFBParameterName = Int32ul

BufferDescriptor = uint64_t

SwapCompleteData = Bytes(0x12)
SwapInfoBlob = Struct(
    "unk" / Bytes(0x6c4),
    Ver("V >= V13_5", "unk_13_3" / Bytes(0x10)),
)

SWAP_SURFACES = 4

Rect = NamedTuple("rect", "x y w h", Int32ul[4])

ActiveRegion = NamedTuple("rect", "max_x max_y min_w max_w", Int32ul[4])

IOMFBSwapRec = Struct(
    "ts1" / Default(Int64ul, 0),
    "ts2" / Default(Int64ul, 0),
    "unk_10" / Default(Int64ul, 0),
    "unk_18" / Default(Int64ul, 0),
    "ts64_unk" / Default(Int64ul, 0),
    "unk_28" / Default(Int64ul, 0),
    "ts3" / Default(Int64ul, 0),
    "unk_38" / Default(Int64ul, 0),
    "flags1" / Hex(Int64ul),
    "flags2" / Hex(Int64ul),
    "swap_id" / Int32ul,
    "surf_ids" / Int32ul[SWAP_SURFACES],
    "src_rect" / Rect[SWAP_SURFACES],
    "surf_flags" / Int32ul[SWAP_SURFACES],
    "surf_unk" / Int32ul[SWAP_SURFACES],
    "dst_rect" / Rect[SWAP_SURFACES],
    "swap_enabled" / Hex(Int32ul),
    "swap_completed" / Hex(Int32ul),
    "bg_color" / Hex(Default(Int32ul, 0)),
    "unk_110" / UnkBytes(0x30),
    "active_region_enable" / Default(Int32ul[SWAP_SURFACES], [0]*SWAP_SURFACES),
    "active_regions" / Default(ActiveRegion[SWAP_SURFACES], [(0,0,0,0)] * SWAP_SURFACES),
    "unk_190" / UnkBytes(0x138),
    "unk_2c8" / Hex(Default(Int32ul, 0)),
    "unk_2cc" / UnkBytes(0x14),
    "unk_2e0" / Hex(Default(Int32ul, 0)),
    Ver("V < V13_5", "unk_2e2" / UnkBytes(0x2)),
    Ver("V >= V13_5", "unk_2e2" / UnkBytes(0x3)),
    "bl_unk" / Hex(Int64ul), # seen: 0x0, 0x1, 0x101, 0x1_0000, 0x101_010101
    "bl_val" / Hex(Int32ul), # range 0x10000000 - approximately 0x7fe07fc0 for 4 - 510 nits
    "bl_power" / Hex(Int8ul), # constant 0x40, 0x00: backlight off
    "unk_2f3" / UnkBytes(0x2d),
    Ver("V >= V13_5", "unk_320" / UnkBytes(0x147)),
)

MAX_PLANES = 3

ComponentTypes = Struct(
    "count" / Int8ul,
    "types" / SizedArray(7, "count", Int8ul),
)

#ComponentTypes = Bytes(8)

CLA_Chroma = 0
CLA_Luma = 1
CLA_Alpha = 2

ARGB_RedLuma = 0
ARGB_GreenCb = 1
ARGB_BlueCr = 2
ARGB_Alpha = 3

kIOSurfaceAddressFormatLinear = 0
kIOSurfaceAddressFormatIndirect = 1
kIOSurfaceAddressFormatTwiddled = 2
kIOSurfaceAddressFormatTiled = 3
kIOSurfaceAddressFormatReference = 4

kIOSurfaceCompressionTypeNone = 0
kIOSurfaceCompressionTypeHTPC = 1
kIOSurfaceCompressionTypeAGX = 2
kIOSurfaceCompressionTypeInterchange = 3
kIOSurfaceCompressionTypeInterchangeLossy = 4

kInterchangeSizeNone = 0
kInterchangeSizeLarge = 1
kInterchangeSizeNormal = 2

kInterchangeLinear = 0
kInterchangeCompressed = 1
kInterchangeUncompressed = 2

PlaneInfo = Struct(
    "width" / Int32ul,
    "height" / Int32ul,
    "base" / Hex(Int32ul),
    "offset" / Hex(Int32ul),
    "stride" / Hex(Int32ul),
    "size" / Hex(Int32ul),
    "tile_size" / Int16ul,
    "tile_w" / Int8ul,
    "tile_h" / Int8ul,
    "unk1" / UnkBytes(0xd),
    "unk2" / Hex(Int8ul),
    "unk3" / UnkBytes(0x26),
)

assert PlaneInfo.sizeof() == 0x50

CompressionInfo = Struct(
    "tile_w" / Int32ul,
    "tile_h" / Int32ul,
    "meta_offset" / Int32ul,
    "data_offset" / Int32ul,
    "tile_meta_bytes" / Int32ul,
    "tiles_w" / Int32ul,
    "tiles_h" / Int32ul,
    "unk1" / Default(Int32ul, 0),
    "compression_type" / Default(Int32ul, 3),
    "unk3" / Default(Int32ul, 0),
    "pad" / Padding(3),
    "tile_bytes" / Int32ul,
    "row_stride" / Int32ul,
    "pad2" / Padding(1),
)

assert CompressionInfo.sizeof() == 0x34

IOSurface = Struct(
    "is_tiled" / bool_,
    "unk_1" / bool_,
    "unk_2" / bool_,
    "plane_cnt" / Int32ul,
    "plane_cnt2" / Int32ul,
    "format" / FourCC,
    "unk_f" / Default(Hex(Int32ul), 0),
    "xfer_func" / Int8ul,
    "colorspace" / Int8ul,
    "stride" / Int32ul,
    "pix_size" / Int16ul,
    "pel_w" / Int8ul,
    "pel_h" / Int8ul,
    "offset" / Default(Hex(Int32ul), 0),
    "width" / Int32ul,
    "height" / Int32ul,
    "buf_size" / Hex(Int32ul),
    "unk_2d" / Default(Int32ul, 0),
    "unk_31" / Default(Int32ul, 0),
    "surface_id" / Int32ul,
    "comp_types" / Default(SizedArray(MAX_PLANES, "plane_cnt", ComponentTypes), []),
    "has_comp" / Bool(Int64ul),
    "planes" / Default(SizedArray(MAX_PLANES, "plane_cnt", PlaneInfo), []),
    "has_planes" / Bool(Int64ul),
    "compression_info" / Default(SizedArray(MAX_PLANES, "plane_cnt", CompressionInfo), []),
    "has_compr_info" / Bool(Int64ul),
    "unk_1f5" / Int32ul,
    "unk_1f9" / Int32ul,
    "padding" / UnkBytes(7),
    Ver("V >= V13_5", "padding_13_3" / UnkBytes(40)),
)

IOMFBColorFixedMatrix = Array(5, Array(3, ulong))

class PropID(IntEnum):
    BrightnessCorrection = 14

class UPPipeAP_H13P(IPCObject):
    # FW version dependent Calls
    if Ver.check("V < V13_5"):
        late_init_signal = Call(bool_, "late_init_signal")
        update_notify_clients_dcp = Call(void, "update_notify_clients_dcp", Array(26, uint))
    else:
        late_init_signal = Call(bool_, "late_init_signal", bool_)
        update_notify_clients_dcp = Call(void, "update_notify_clients_dcp", Array(26, uint))

    A000 = late_init_signal
    A029 = Call(void, "setup_video_limits")
    A034 = update_notify_clients_dcp
    A035 = Call(bool_, "is_hilo")
    A036 = Call(bool_, "apt_supported")
    A037 = Call(uint, "get_dfb_info", InOutPtr(uint), InOutPtr(Array(4, ulong)), InOutPtr(uint))
    A038 = Call(uint, "get_dfb_compression_info", InOutPtr(uint))

    D000 = Callback(bool_, "did_boot_signal")
    D001 = Callback(bool_, "did_power_on_signal")
    D002 = Callback(void, "will_power_off_signal")
    D003 = Callback(void, "rt_bandwidth_setup_ap", config=OutPtr(rt_bw_config_t))
    D006 = Callback(void, "set_frame_sync_props", props=InOutPtr(frame_sync_props_t))

IdleCachingState = uint32_t

class UnifiedPipeline2(IPCObject):
    # FW version dependent Call tags
    set_create_DFB = Call(void, "set_create_DFB")
    vi_set_temperature_hint = Call(IOMFBStatus, "vi_set_temperature_hint")

    if Ver.check("V < V13_5"):
        A357 = set_create_DFB
        A358 = vi_set_temperature_hint
    else:
        A373 = set_create_DFB
        A374 = vi_set_temperature_hint

    A352 = Call(bool_, "applyProperty", uint, uint)
    A353 = Call(uint, "get_system_type")

    D100 = Callback(void, "match_pmu_service")
    D101 = Callback(uint32_t, "UNK_get_some_field")
    D102 = Callback(void, "set_number_property", key=string(0x40), value=uint)

    # FW version dependent Callback tags

    cb_set_boolean_property = Callback(void, "set_boolean_property", key=string(0x40), value=bool_)
    cb_removeProperty = Callback(void, "removeProperty", key=string(0x40))
    cb_create_provider_service = Callback(bool_, "create_provider_service")
    cb_create_product_service = Callback(bool_, "create_product_service")
    cb_create_PMU_service = Callback(bool_, "create_PMU_service")
    cb_create_iomfb_service = Callback(bool_, "create_iomfb_service")
    cb_create_backlight_service = Callback(bool_, "create_backlight_service")

    cb_create_nvram_service = Callback(bool_, "create_nvram_service")
    cb_set_idle_caching_state_ap = Callback(void, "set_idle_caching_state_ap", IdleCachingState, uint)
    cb_start_hardware_boot = Callback(bool_, "start_hardware_boot")
    cb_is_dark_boot = Callback(bool_, "is_dark_boot")
    cb_is_waking_from_hibernate = Callback(bool_, "is_waking_from_hibernate")
    cb_read_edt_data = Callback(bool_, "read_edt_data", key=string(0x40), count=uint, value=InOut(Lazy(SizedArray(8, "count", uint32_t))))
    cb_setDCPAVPropStart = Callback(bool_, "setDCPAVPropStart", length=uint)
    cb_setDCPAVPropChunk = Callback(bool_, "setDCPAVPropChunk", data=HexDump(SizedBytes(0x1000, "length")), offset=uint, length=uint)
    cb_setDCPAVPropEnd = Callback(bool_, "setDCPAVPropEnd", key=string(0x40))
    cb_allocate_bandwidth = Callback(bool_, "allocate_badwidth", InOutPtr(ulong), InOutPtr(ulong), ulong)

    if Ver.check("V < V13_5"):
        D103 = cb_set_boolean_property
        D106 = cb_removeProperty
        D107 = cb_create_provider_service
        D108 = cb_create_product_service
        D109 = cb_create_PMU_service
        D110 = cb_create_iomfb_service
        D111 = cb_create_backlight_service
        D112 = cb_set_idle_caching_state_ap
        D116 = cb_start_hardware_boot
        D117 = cb_is_dark_boot
        D118 = cb_is_waking_from_hibernate
        D120 = cb_read_edt_data
        D122 = cb_setDCPAVPropStart
        D123 = cb_setDCPAVPropChunk
        D124 = cb_setDCPAVPropEnd
    else:
        D103 = Callback(void, "trigger_user_cal_loader")
        D104 = cb_set_boolean_property
        D107 = cb_removeProperty
        D108 = cb_create_provider_service
        D109 = cb_create_product_service
        D110 = cb_create_PMU_service
        D111 = cb_create_iomfb_service
        D112 = cb_create_backlight_service
        D113 = cb_create_nvram_service
        D114 = Callback(bool_, "get_tiling_state", event=uint, para=uint, val=InOutPtr(uint))
        D115 = Callback(bool_, "set_tiling_state", event=uint, para=uint, val=InPtr(uint))
        D116 = cb_set_idle_caching_state_ap
        D120 = cb_start_hardware_boot
        D121 = cb_is_dark_boot
        D122 = cb_is_waking_from_hibernate
        D124 = cb_read_edt_data
        D126 = cb_setDCPAVPropStart
        D127 = cb_setDCPAVPropChunk
        D128 = cb_setDCPAVPropEnd
        D129 = cb_allocate_bandwidth

class UPPipe2(IPCObject):
    A102 = Call(uint64_t, "test_control", cmd=uint64_t, arg=uint)
    A103 = Call(void, "get_config_frame_size", width=InOutPtr(uint), height=InOutPtr(uint))
    A104 = Call(void, "set_config_frame_size", width=uint, height=uint)
    A105 = Call(void, "program_config_frame_size")
    A130 = Call(bool_, "init_ca_pmu")
    A131 = Call(bool_, "pmu_service_matched")
    A132 = Call(bool_, "backlight_service_matched")

    # FW version dependent Callback tags
    cb_get_calendar_time_ms = Callback(uint64_t, "get_calendar_time_ms")
    cb_update_backlight_factor_prop = Callback(void, "update_backlight_factor_prop", int_)

    D201 = Callback(uint32_t, "map_buf", buf=InPtr(BufferDescriptor), vaddr=OutPtr(ulong), dva=OutPtr(ulong), unk=bool_)
    D202 = Callback(void, "unmap_buf", buf=InPtr(BufferDescriptor), unk1=uint, unk2=ulong, unkB=uint)

    D206 = Callback(bool_, "match_pmu_service_2")
    D207 = Callback(bool_, "match_backlight_service")

    if Ver.check("V < V13_5"):
        D208 = cb_get_calendar_time_ms
        D211 = cb_update_backlight_factor_prop
    else:
        D208 = cb_update_backlight_factor_prop
        D209 = cb_get_calendar_time_ms

class PropRelay(IPCObject):
    if Ver.check("V < V13_5"):
        D300 = Callback(void, "pr_publish", prop_id=uint32_t, value=int_)
    else:
        D300 = Callback(void, "pr_publish", prop_id=uint32_t, value=int_, unk0=int_, unk1=int_)

class IOMobileFramebufferAP(IPCObject):
    # FW version dependent Calls
    if Ver.check("V < V13_5"):
        swap_submit_dcp = Call(uint32_t, "swap_submit_dcp",
                 swap_rec=InPtr(IOMFBSwapRec),
                 surfaces=Array(4, InPtr(IOSurface)),
                 surfAddr=Array(4, Hex(ulong)),
                 unkBool=bool_,
                 unkFloat=Float64l,
                 unkInt=uint,
                 unkOutBool=OutPtr(bool_))
    else:
        swap_submit_dcp = Call(uint32_t, "swap_submit_dcp",
                swap_rec=InPtr(IOMFBSwapRec),
                surfaces=Array(SWAP_SURFACES, InPtr(IOSurface)),
                surfAddr=Array(SWAP_SURFACES, Hex(ulong)),
                unkU64Array=Array(SWAP_SURFACES, Hex(ulong)),
                surfaces2=Array(5, InPtr(IOSurface)),
                surfAddr2=Array(5, Hex(ulong)),
                unkBool=bool_,
                unkFloat=Float64l,
                unkU64=ulong,
                unkBool2=bool_,
                unkInt=uint,
                unkOutBool=OutPtr(bool_),
                unkCUintArray=InPtr(uint),
                unkUintPtr=OutPtr(uint))

    A401 = Call(uint32_t, "start_signal")
    A407 = Call(uint32_t, "swap_start", swap_id=InOutPtr(uint), client=InOutPtr(IOUserClient))
    A408 = swap_submit_dcp
    A410 = Call(uint32_t, "set_display_device", uint)
    A411 = Call(bool_, "is_main_display")

    A412 = Call(uint32_t, "set_digital_out_mode", uint, uint)
    A413 = Call(uint32_t, "get_digital_out_state", InOutPtr(uint))
    A414 = Call(uint32_t, "get_display_area", InOutPtr(ulong))
    A419 = Call(uint32_t, "get_gamma_table", InOutPtr(Bytes(0xc0c)))
    A422 = Call(uint32_t, "set_matrix", uint, InPtr(Array(3, Array(3, ulong))))
    A423 = Call(uint32_t, "set_contrast", InOutPtr(Float32l))
    A426 = Call(uint32_t, "get_color_remap_mode", InOutPtr(uint32_t))
    A427 = Call(uint32_t, "setBrightnessCorrection", uint)

    # FW version dependent Call tags
    set_block_dcp = Call(uint32_t, "set_block_dcp", arg1=uint64_t, arg2=uint, arg3=uint, arg4=Array(8, ulong), arg5=uint, data=SizedBytes(0x1000, "length"), length=ulong, unknArry=Array(4, uint))
    get_block_dcp = Call(uint32_t, "get_block_dcp", arg1=uint64_t, arg2=uint, arg3=uint, arg4=Array(8, ulong), arg5=uint, data=OutPtr(SizedBytes(0x1000, "length")), length=uint)
    swap_set_color_matrix = Call(uint32_t, "swap_set_color_matrix", matrix=InOutPtr(IOMFBColorFixedMatrix), func=uint32_t, unk=uint)
    set_parameter_dcp = Call(uint32_t, "set_parameter_dcp", param=IOMFBParameterName, value=Lazy(SizedArray(4, "count", ulong)), count=uint)
    display_width = Call(uint, "display_width")
    display_height = Call(uint, "display_height")
    get_display_size = Call(void, "get_display_size", OutPtr(uint), OutPtr(uint))
    do_create_default_frame_buffer = Call(int_, "do_create_default_frame_buffer")
    printRegs = Call(void, "printRegs")
    enable_disable_video_power_savings = Call(int_, "enable_disable_video_power_savings", uint)
    first_client_open = Call(void, "first_client_open")
    last_client_close_dcp = Call(void, "last_client_close_dcp", OutPtr(uint))
    writeDebugInfo = Call(bool_, "writeDebugInfo", ulong)
    flush_debug_flags = Call(void, "flush_debug_flags", uint)
    io_fence_notify = Call(bool_, "io_fence_notify", uint, uint, ulong, IOMFBStatus)
    setDisplayRefreshProperties = Call(bool_, "setDisplayRefreshProperties")
    flush_supportsPower = Call(void, "flush_supportsPower", bool_)
    abort_swaps_dcp = Call(uint, "abort_swaps_dcp", InOutPtr(IOMobileFramebufferUserClient))
    update_dfb = Call(uint, "update_dfb", surf=InPtr(IOSurface))
    setPowerState = Call(uint32_t, "setPowerState", ulong, bool_, OutPtr(uint))
    isKeepOnScreen = Call(bool_, "isKeepOnScreen")

    if Ver.check("V < V13_5"):
        A435 = set_block_dcp
        A436 = get_block_dcp
        A438 = swap_set_color_matrix
        A439 = set_parameter_dcp
        A440 = display_width
        A441 = display_height
        A442 = get_display_size
        A443 = do_create_default_frame_buffer
        A444 = printRegs
        A447 = enable_disable_video_power_savings
        A454 = first_client_open
        A455 = last_client_close_dcp
        A456 = writeDebugInfo
        A457 = flush_debug_flags
        A458 = io_fence_notify
        A460 = setDisplayRefreshProperties
        A463 = flush_supportsPower
        A464 = abort_swaps_dcp
        A467 = update_dfb
        A468 = setPowerState
        A469 = isKeepOnScreen
    else:
        A437 = set_block_dcp
        A438 = get_block_dcp
        A440 = swap_set_color_matrix
        A441 = set_parameter_dcp
        A442 = display_width
        A443 = display_height
        A444 = get_display_size
        A445 = do_create_default_frame_buffer
        A446 = printRegs
        A449 = enable_disable_video_power_savings
        A456 = first_client_open
        A457 = last_client_close_dcp
        A458 = writeDebugInfo
        A459 = flush_debug_flags
        A460 = io_fence_notify
        A463 = setDisplayRefreshProperties
        A466 = flush_supportsPower
        A467 = abort_swaps_dcp
        A468 = Call(uint, "remove_gain_maps", InOutPtr(IOMobileFramebufferUserClient))
        A470 = update_dfb
        A472 = setPowerState
        A473 = isKeepOnScreen

    # FW version dependent callbacks
    if Ver.check("V < V13_5"):
        hotPlug_notify_gated = Callback(void, "hotPlug_notify_gated", ulong)
    else:
        # TODO: is this sensible?
        hotPlug_notify_gated = Callback(void, "hotPlug_notify_gated", uint, InOutPtr(Bytes(0x4c)))

    D552 = Callback(bool_, "setProperty_dict", key=string(0x40), value=InPtr(Padded(0x1000, OSDictionary())))
    D561 = Callback(bool_, "setProperty_dict", key=string(0x40), value=InPtr(Padded(0x1000, OSDictionary())))
    D563 = Callback(bool_, "setProperty_int", key=string(0x40), value=InPtr(uint64_t))
    D565 = Callback(bool_, "setProperty_bool", key=string(0x40), value=InPtr(Bool(uint32_t)))
    D567 = Callback(bool_, "setProperty_str", key=string(0x40), value=string(0x40))

    D574 = Callback(IOMFBStatus, "powerUpDART", bool_)

    D575 = Callback(bool_, "get_dot_pitch", OutPtr(uint))
    D576 = hotPlug_notify_gated
    D577 = Callback(void, "powerstate_notify", bool_, bool_)
    D578 = Callback(bool_, "idle_fence_create", IdleCachingState)
    D579 = Callback(void, "idle_fence_complete")

    D581 = Callback(void, "swap_complete_head_of_line", uint, bool_, uint, bool_)
    D582 = Callback(bool_, "create_default_fb_surface", uint, uint)
    D583 = Callback(bool_, "serializeDebugInfoCb", ulong, InPtr(uint64_t), uint)
    D584 = Callback(void, "clear_default_surface")

    D588 = Callback(void, "resize_default_fb_surface_gated")
    D589 = Callback(void, "swap_complete_ap_gated", swap_id=uint, unkBool=bool_, swap_data=InPtr(SwapCompleteData), swap_info=SwapInfoBlob, unkUint=uint)

    D591 = Callback(void, "swap_complete_intent_gated", swap_id=uint, unkB=bool_, unkInt=uint32_t, width=uint, height=uint)
    D592 = Callback(void, "abort_swap_ap_gated", swap_id=uint)
    D593 = Callback(void, "enable_backlight_message_ap_gated", bool_)
    D594 = Callback(void, "setSystemConsoleMode", bool_)

    D596 = Callback(bool_, "isDFBAllocated")
    D597 = Callback(bool_, "preserveContents")
    D598 = Callback(void, "find_swap_function_gated")

class ServiceRelay(IPCObject):
    # FW version dependent Callbacks
    if Ver.check("V < V13_5"):
        sr_mapDeviceMemoryWithIndex = Callback(IOMFBStatus, "sr_mapDeviceMemoryWithIndex", obj=FourCC, index=uint, flags=uint, addr=OutPtr(ulong), length=OutPtr(ulong))
    else:
        sr_mapDeviceMemoryWithIndex = Callback(IOMFBStatus, "sr_mapDeviceMemoryWithIndex", obj=FourCC, index=uint, flags=uint, unk_u64=OutPtr(ulong), addr=OutPtr(ulong), length=OutPtr(ulong))

    D400 = Callback(void, "get_property", obj=FourCC, key=string(0x40), value=OutPtr(Bytes(0x200)), length=InOutPtr(uint))
    D401 = Callback(bool_, "sr_get_uint_prop", obj=FourCC, key=string(0x40), value=InOutPtr(ulong))
    D404 = Callback(void, "sr_set_uint_prop", obj=FourCC, key=string(0x40), value=uint)
    D406 = Callback(void, "set_fx_prop", obj=FourCC, key=string(0x40), value=uint)
    D408 = Callback(uint64_t, "sr_getClockFrequency", obj=FourCC, arg=uint)
    D411 = sr_mapDeviceMemoryWithIndex
    D413 = Callback(bool_, "sr_setProperty_dict", obj=FourCC, key=string(0x40), value=InPtr(Padded(0x1000, OSDictionary())))
    D414 = Callback(bool_, "sr_setProperty_int", obj=FourCC, key=string(0x40), value=InPtr(uint64_t))
    D415 = Callback(bool_, "sr_setProperty_bool", obj=FourCC, key=string(0x40), value=InPtr(Bool(uint32_t)))

mem_desc_id = uint

class MemDescRelay(IPCObject):
    D451 = Callback(mem_desc_id, "allocate_buffer", uint, ulong, uint, OutPtr(ulong), OutPtr(ulong), OutPtr(ulong))
    D452 = Callback(mem_desc_id, "map_physical", paddr=ulong, size=ulong, flags=uint, dva=OutPtr(ulong), dvasize=OutPtr(ulong))
    D453 = Callback(mem_desc_id, "withAddressRange", ulong, ulong, uint, uint64_t, OutPtr(uint), OutPtr(ulong))
    D454 = Callback(IOMFBStatus, "prepare", uint, uint)
    D455 = Callback(IOMFBStatus, "complete", uint, uint)
    D456 = Callback(bool_, "release_descriptor", uint)

ALL_CLASSES = [
    UPPipeAP_H13P,
    UnifiedPipeline2,
    IOMobileFramebufferAP,
    ServiceRelay,
    PropRelay,
    UPPipe2,
    MemDescRelay,
]

ALL_METHODS = {}

for cls in ALL_CLASSES:
    ALL_METHODS.update(cls.methods())

SHORT_CHANNELS = {
    "CB": "d",
    "CMD": "C",
    "ASYNC": "a",
    "OOBCMD": "O",
    "OOBCB": "o",
}

RDIR = { ">": "<", "<": ">" }

class Call:
    def __init__(self, dir, chan, off, msg, in_size, out_size, in_data=b''):
        self.dir = dir
        self.chan = chan
        self.msg = msg
        self.off = off
        self.in_size = in_size
        self.out_size = out_size
        self.in_data = in_data
        self.out_data = None
        self.complete = False
        self.ret = None

    def ack(self, out_data):
        self.out_data = out_data
        self.complete = True

    def print_req(self, indent=""):
        log = f"{indent}{self.dir}{SHORT_CHANNELS[self.chan]}[{self.off:#x}] {self.msg} "

        cls, method = ALL_METHODS.get(self.msg, (None, None))
        if cls is None:
            print(log + f"unknown: {self.in_size:#x}/{self.out_size:#x}")
            return

        log += f"{cls.__name__}::{method.name}("
        in_size = method.in_struct.sizeof()

        if in_size != len(self.in_data):
            print(f"{log} !! Expected {in_size:#x} bytes, got {len(self.in_data):#x} bytes (in)")
            dump_fields(method.in_fields)
            chexdump(self.in_data)
            self.in_vals = {}
            return

        self.in_vals = method.parse_input(self.in_data)

        log += f"{method.fmt_args(self.in_vals)})"

        print(log)

        method.print_long_args(indent, self.in_vals)
        #if method.in_fields:
            #print(self.in_vals)

    def print_reply(self, indent=""):
        assert self.complete
        log = f"{indent}{RDIR[self.dir]}{SHORT_CHANNELS[self.chan]}[{self.off:#x}] {self.msg} "

        cls, method = ALL_METHODS.get(self.msg, (None, None))
        if cls is None:
            print(log + f"{self.in_size:#x}/{self.out_size:#x}")
            return

        log += f"{cls.__name__}::{method.name}("
        out_size = method.out_struct.sizeof()

        if out_size != len(self.out_data):
            print(f"{log} !! Expected {out_size:#x} bytes, got {len(self.out_data):#x} bytes (out)")
            dump_fields(method.out_fields)
            chexdump(self.out_data)
            return

        self.out_vals = method.parse_output(self.out_data, self.in_vals)

        log += f"{method.fmt_args(self.in_vals, self.out_vals)})"

        if "ret" in self.out_vals:
            self.ret = self.out_vals.ret
            del self.out_vals["ret"]
            log += f" = {self.ret!r}"

        print(log)

        method.print_long_args(indent, self.in_vals, self.out_vals)
        #if len(method.out_fields) - (self.ret is not None):
            #print(self.out_vals)

    def get_method(self):
        cls, method = ALL_METHODS.get(self.msg, (None, None))
        return method



# SPDX-License-Identifier: MIT

import struct
from construct import *

from m1n1.hv import TraceMode
from m1n1.proxyutils import RegMonitor
from m1n1.utils import *

from m1n1.trace import ADTDevTracer
from m1n1.trace.asc import ASCTracer, ASCRegs, EP, EPState, msg, msg_log, DIR
from m1n1.trace.dart import DARTTracer
from m1n1.trace.gpio import GPIOTracer
from m1n1.trace.spi import SPITracer

DARTTracer = DARTTracer._reloadcls()
ASCTracer = ASCTracer._reloadcls()
GPIOTracer = GPIOTracer._reloadcls()
SPITracer = SPITracer._reloadcls()

# SPI HID transport tracer for 2021 macbook models

kbd_node = None
for node in hv.adt.walk_tree():
    try:
        if node.compatible[0] == "spi-1,spimc":
            for c in node:
                try:
                    if c.compatible[0] == "hid-transport,spi":
                        kbd_node = c
                        break
                except AttributeError:
                    continue
    except AttributeError:
        continue
    if kbd_node is not None:
        break

# trace interrupts
aic_phandle = getattr(hv.adt["/arm-io/aic"], "AAPL,phandle")
spi_node = kbd_node._parent

if getattr(spi_node, "interrupt-parent") == aic_phandle:
    for irq in getattr(spi_node, "interrupts"):
        hv.trace_irq(node.name, irq, 1, hv.IRQTRACE_IRQ)

spi_pins = {
    0x37: "spi3_cs",
    0xc2: "ipd_en",
}
spi_pins_nub = {
    0x6: "ipd_irq",
}

SPI_HID_PKT = Struct(
    "flags" / Int8ub,
    "dev"   / Int8ub,
    "offset"    / Int16ul,
    "remain"    / Int16ul,
    "length"    / Int16ul,
    "data"      / Bytes(246),
    "crc16"     / Int16ul
)

#gpio_tracer = GPIOTracer(hv, "/arm-io/gpio0", spi_pins, verbose=0)
#gpio_tracer.start()

#gpio_tracer_nub = GPIOTracer(hv, "/arm-io/nub-gpio0", spi_pins_nub, verbose=0)
#gpio_tracer_nub.start()

dart_sio_tracer = DARTTracer(hv, "/arm-io/dart-sio", verbose=1)
dart_sio_tracer.start()

iomon = RegMonitor(hv.u, ascii=True)

def readmem_iova(addr, size):
    try:
        return dart_sio_tracer.dart.ioread(0, addr, size)
    except Exception as e:
        print(e)
        return None

iomon.readmem = readmem_iova

class SIOMessage(Register64):
    EP    = 7, 0 # matches device's ADT dma-channels, j314c spi3 0x1a and 0x1b
    TAG   = 13, 8 # counts for ipd spi transfers from 0x02 to 0x20
    TYPE  = 23, 16 # OP
    PARAM = 31, 24
    DATA  = 63, 32

class SIOStart(SIOMessage):
    TYPE = 23, 16, Constant(2)

class SIOSetup(SIOMessage):
    TYPE = 23, 16, Constant(3)

class SIOConfig(SIOMessage): #???
    TYPE = 23, 16, Constant(5)

class SIOAck(SIOMessage):
    TYPE = 23, 16, Constant(0x65)

class SIOSetupIO(SIOMessage):
    TYPE = 23, 16, Constant(6)

class SIOCompleteIO(SIOMessage):
    TYPE = 23, 16, Constant(0x68)

class SIOEp(EP):
    BASE_MESSAGE = SIOMessage
    SHORT = "sioep"

    def __init__(self, tracer, epid):
        super().__init__(tracer, epid)
        self.state.iova = None
        self.state.iova_cfg = None
        self.state.dumpfile = None

    @msg(2, DIR.TX, SIOStart)
    def Start(self, msg):
        self.log("Start SIO")

    @msg(3, DIR.TX, SIOSetup)
    def m_Setup(self, msg):
        iomon.poll()
        if msg.EP == 0 and msg.PARAM == 0x1:
            self.state.iova = msg.DATA << 12
        elif msg.EP == 0 and msg.PARAM == 0x2:
            # size for PARAM == 0x1?
            if self.state.iova is not None and self.tracer.verbose > 1:
                iomon.add(self.state.iova, msg.DATA * 8, name=f"sio.shmem@{self.state.iova:08x}",
                          offset=self.state.iova)
        elif msg.EP == 0 and msg.PARAM == 0xb:
            # second iova block, maybe config
            self.state.iova_cfg = msg.DATA << 12
        elif msg.EP == 0 and msg.PARAM == 0xc:
            # size for PARAM == 0xb?
            if self.state.iova is not None and self.tracer.verbose > 1:
                iomon.add(self.state.iova_cfg, msg.DATA * 8,
                          name=f"sio.shmem@{self.state.iova_cfg:08x}", offset=self.state.iova_cfg)

    @msg(0x65, DIR.RX, SIOAck)
    def m_Ack(self, msg):
        iomon.poll()

    @msg(6, DIR.TX, SIOSetupIO)
    def m_SetupIO(self, msg):
        iomon.poll()
        if self.state.iova is None:
            return
        if msg.EP == 0x1a:
            buf = struct.unpack("<I", self.tracer.ioread(self.state.iova + 0xa8, 4))[0]
            size = struct.unpack("<I", self.tracer.ioread(self.state.iova + 0xb0, 4))[0]
            self.log_spihid_pkt("SPI3 TX", self.tracer.ioread(buf, size))

    @msg(0x68, DIR.RX, SIOCompleteIO)
    def m_CompleteIO(self, msg):
        iomon.poll()
        if self.state.iova is None:
            return
        if msg.EP == 0x1b:
            buf = struct.unpack("<I", self.tracer.ioread(self.state.iova + 0x48, 4))[0]
            size = struct.unpack("<I", self.tracer.ioread(self.state.iova + 0x50, 4))[0]
            self.log_spihid_pkt("SPI3 RX", self.tracer.ioread(buf, size))

    def log_spihid_pkt(self, label, data):
        if len(data) != 256:
            self.log(f"{label}: unexpected data length: {len(data):d}")
            chexdump(data)
            return

        crc16 = crc16USB(0, data[:254])
        pkt = SPI_HID_PKT.parse(data)
        #self.log(f"pkt.crc16:{pkt.crc16:#04x} crc16:{crc16:#04x}")
        if pkt.length == 0:
            return

        if pkt.flags == 0x80 and pkt.dev == 0x11 and pkt.length == 849 and pkt.offset == 256 and pkt.remain == 834 and crc16 == 0x1489:
            return
        if pkt.crc16 != crc16:
            self.log(f"{label}: CRC mismatch: pkt.crc16:{pkt.crc16:#04x} crc16:{crc16:#04x}")
            chexdump(data)
            return
        self.log(f"{label}: flags:{pkt.flags:#2x} dev:{pkt.dev:#2x} length:{pkt.length:4d} offset:{pkt.offset:3d} remain:{pkt.remain:3d}")
        chexdump(pkt.data[:min(246, pkt.length)])

        if self.state.dumpfile:
            dump = f"{label}: flags:{pkt.flags:#2x} dev:{pkt.dev:#2x} length:{pkt.length:4d}  {pkt.data[:min(246, pkt.length)].hex()}\n"
            self.state.dumpfile.write(dump)
            self.state.dumpfile.flush()


class SIOTracer(ASCTracer):
    ENDPOINTS = {
        0x20: SIOEp
    }


sio_tracer = SIOTracer(hv, "/arm-io/sio", verbose=False)
sio_tracer.start(dart_sio_tracer.dart)

sio_tracer.ep.sioep.state.dumpfile = open("spi_hid.log", "a")

spi_tracer = SPITracer(hv, "/arm-io/" + spi_node.name, verbose=1)
spi_tracer.start()



# SPDX-License-Identifier: MIT
import platform, os, sys, struct, serial, time
from construct import *
from enum import IntEnum, IntFlag
from serial.tools.miniterm import Miniterm

from .utils import *
from .sysreg import *

__all__ = ["REGION_RWX_EL0", "REGION_RW_EL0", "REGION_RX_EL1"]

# Hack to disable input buffer flushing
class Serial(serial.Serial):
    def _reset_input_buffer(self):
        return

    def reset_input_buffer(self):
        return

class UartError(RuntimeError):
    pass

class UartTimeout(UartError):
    pass

class UartCMDError(UartError):
    pass

class UartChecksumError(UartError):
    pass

class UartRemoteError(UartError):
    pass

class Feature(IntFlag):
    DISABLE_DATA_CSUMS = 0x01  # Data transfers don't use checksums

    @classmethod
    def get_all(cls):
        return cls.DISABLE_DATA_CSUMS

    def __str__(self):
        return ", ".join(feature.name for feature in self.__class__
            if feature & self) or "<none>"


class START(IntEnum):
    BOOT = 0
    EXCEPTION = 1
    EXCEPTION_LOWER = 2
    HV = 3

class EXC(IntEnum):
    SYNC = 0
    IRQ = 1
    FIQ = 2
    SERROR = 3

class EVENT(IntEnum):
    MMIOTRACE = 1
    IRQTRACE = 2

class EXC_RET(IntEnum):
    UNHANDLED = 1
    HANDLED = 2
    EXIT_GUEST = 3
    STEP = 4

class DCP_SHUTDOWN_MODE(IntEnum):
    QUIESCED = 0
    SLEEP_IF_EXTERNAL = 1
    SLEEP = 2

class PIX_FMT(IntEnum):
    XRGB = 0
    XBGR = 1

class DART(IntEnum):
    T8020 = 0
    T8110 = 1
    T6000 = 2

ExcInfo = Struct(
    "regs" / Array(32, Int64ul),
    "spsr" / RegAdapter(SPSR),
    "elr" / Int64ul,
    "esr" / RegAdapter(ESR),
    "far" / Int64ul,
    "afsr1" / Int64ul,
    "sp" / Array(3, Int64ul),
    "cpu_id" / Int64ul,
    "mpidr" / Int64ul,
    "elr_phys" / Int64ul,
    "far_phys" / Int64ul,
    "sp_phys" / Int64ul,
    "data" / Int64ul,
)
# Sends 56+ byte Commands and Expects 36 Byte Responses
# Commands are format <I48sI
#   4 byte command, 48 byte null padded data + 4 byte checksum
# Responses are of the format: struct format <Ii24sI
#   4byte Response , 4 byte status, 24 byte string,  4 byte Checksum
#    Response must start 0xff55aaXX where XX distinguishes between them
#    In little endian mode these numbers as listed as REQ_* constants
# defined under UartInterface
#
#  Event Response REQ_EVENT passed to registered Event Handler
#  Boot Response REQ_BOOT passed to handle_boot() which may
#       pass to a matching registered handler based on reason, code values
#  If the status is ST_OK returns the data field to caller
#     Otherwise reports a remote Error

class UartInterface(Reloadable):
    REQ_NOP = 0x00AA55FF
    REQ_PROXY = 0x01AA55FF
    REQ_MEMREAD = 0x02AA55FF
    REQ_MEMWRITE = 0x03AA55FF
    REQ_BOOT = 0x04AA55FF
    REQ_EVENT = 0x05AA55FF

    CHECKSUM_SENTINEL = 0xD0DECADE
    DATA_END_SENTINEL = 0xB0CACC10

    ST_OK = 0
    ST_BADCMD = -1
    ST_INVAL = -2
    ST_XFERERR = -3
    ST_CSUMERR = -4

    CMD_LEN = 56
    REPLY_LEN = 36
    EVENT_HDR_LEN = 8

    DEFAULT_UART_DEV="/dev/m1n1"
    DEFAULT_BAUD_RATE=115200
    if platform.system() == 'Darwin':
        DEFAULT_UART_DEV="/dev/cu.usbmodemP_01"

    def __init__(self, device=None, debug=False):
        self.debug = debug
        self.devpath = None
        if device is None:
            device = os.environ.get("M1N1DEVICE", self.DEFAULT_UART_DEV)
        if isinstance(device, str):
            baud = self.DEFAULT_BAUD_RATE
            if ":" in device:
                device, baud = device.rsplit(":", 1)
                baud = int(baud)
            self.devpath = device
            self.baudrate = baud

            device = Serial(self.devpath, baud)

        self.dev = device
        self.dev.timeout = 0
        self.dev.flushOutput()
        self.dev.flushInput()
        self.pted = False
        #d = self.dev.read(1)
        #while d != "":
            #d = self.dev.read(1)
        self.dev.timeout = int(os.environ.get("M1N1TIMEOUT", "3"))
        self.tty_enable = True
        self.handlers = {}
        self.evt_handlers = {}
        self.enabled_features = Feature(0)

    def checksum(self, data):
        sum = 0xDEADBEEF;
        for c in data:
            sum *= 31337
            sum += c ^ 0x5a
            sum &= 0xFFFFFFFF

        return (sum ^ 0xADDEDBAD) & 0xFFFFFFFF

    def data_checksum(self, data):
        if self.enabled_features & Feature.DISABLE_DATA_CSUMS:
            return self.CHECKSUM_SENTINEL

        return self.checksum(data)

    def readfull(self, size):
        d = b''
        while len(d) < size:
            block = self.dev.read(size - len(d))
            if not block:
                raise UartTimeout("Expected %d bytes, got %d bytes"%(size,len(d)))
            d += block
        return d

    def cmd(self, cmd, payload=b""):
        if len(payload) > self.CMD_LEN:
            raise ValueError("Incorrect payload size %d"%len(payload))

        payload = payload.ljust(self.CMD_LEN, b"\x00")
        command = struct.pack("<I", cmd) + payload
        command += struct.pack("<I", self.checksum(command))
        if self.debug:
            print("<<", hexdump(command))
        self.dev.write(command)

    def unkhandler(self, s):
        if not self.tty_enable:
            return
        for c in s:
            if not self.pted:
                sys.stdout.write("TTY> ")
                self.pted = True
            if c == 10:
                self.pted = False
            sys.stdout.write(chr(c))
            sys.stdout.flush()

    def ttymode(self, dev=None):
        if dev is None:
            dev = self.dev

        tout = dev.timeout
        self.tty_enable = True
        dev.timeout = None

        term = Miniterm(dev, eol='cr')
        term.exit_character = chr(0x1d)  # GS/CTRL+]
        term.menu_character = chr(0x14)  # Menu: CTRL+T
        term.raw = True
        term.set_rx_encoding('UTF-8')
        term.set_tx_encoding('UTF-8')

        print('--- TTY mode | Quit: CTRL+] | Menu: CTRL+T ---')
        term.start()
        try:
            term.join(True)
        except KeyboardInterrupt:
            pass

        print('--- Exit TTY mode ---')
        term.join()
        term.close()

        dev.timeout = tout
        self.tty_enable = False

    def reply(self, cmd):
        reply = b''
        while True:
            if not reply or reply[-1] != 255:
                reply = b''
                reply += self.readfull(1)
                if reply != b"\xff":
                    self.unkhandler(reply)
                    continue
            else:
                reply = b'\xff'
            reply += self.readfull(1)
            if reply != b"\xff\x55":
                self.unkhandler(reply)
                continue
            reply += self.readfull(1)
            if reply != b"\xff\x55\xaa":
                self.unkhandler(reply)
                continue
            reply += self.readfull(1)
            cmdin = struct.unpack("<I", reply)[0]
            if cmdin == self.REQ_EVENT:
                reply += self.readfull(self.EVENT_HDR_LEN - 4)
                data_len, event_type = struct.unpack("<HH", reply[4:])
                reply += self.readfull(data_len + 4)
                if self.debug:
                    print(">>", hexdump(reply))
                checksum = struct.unpack("<I", reply[-4:])[0]
                ccsum = self.data_checksum(reply[:-4])
                if checksum != ccsum:
                    print("Event checksum error: Expected 0x%08x, got 0x%08x"%(checksum, ccsum))
                    raise UartChecksumError()
                self.handle_event(EVENT(event_type), reply[self.EVENT_HDR_LEN:-4])
                reply = b''
                continue

            reply += self.readfull(self.REPLY_LEN - 4)
            if self.debug:
                print(">>", hexdump(reply))
            status, data, checksum = struct.unpack("<i24sI", reply[4:])
            ccsum = self.checksum(reply[:-4])
            if checksum != ccsum:
                print("Reply checksum error: Expected 0x%08x, got 0x%08x"%(checksum, ccsum))
                raise UartChecksumError()

            if cmdin != cmd:
                if cmdin == self.REQ_BOOT and status == self.ST_OK:
                    self.handle_boot(data)
                    reply = b''
                    continue
                raise UartCMDError("Reply command mismatch: Expected 0x%08x, got 0x%08x"%(cmd, cmdin))
            if status != self.ST_OK:
                if status == self.ST_BADCMD:
                    raise UartRemoteError("Reply error: Bad Command")
                elif status == self.ST_INVAL:
                    raise UartRemoteError("Reply error: Invalid argument")
                elif status == self.ST_XFERERR:
                    raise UartRemoteError("Reply error: Data transfer failed")
                elif status == self.ST_CSUMERR:
                    raise UartRemoteError("Reply error: Data checksum failed")
                else:
                    raise UartRemoteError("Reply error: Unknown error (%d)"%status)
            return data

    def handle_boot(self, data):
        reason, code, info = struct.unpack("<IIQ", data[:16])
        reason = START(reason)
        if reason in (START.EXCEPTION, START.EXCEPTION_LOWER):
            code = EXC(code)
        if (reason, code) in self.handlers:
            self.handlers[(reason, code)](reason, code, info)
        elif reason != START.BOOT:
            print(f"Proxy callback without handler: {reason}, {code}")

    def set_handler(self, reason, code, handler):
        self.handlers[(reason, code)] = handler

    def handle_event(self, event_id, data):
        if event_id in self.evt_handlers:
            self.evt_handlers[event_id](data)

    def set_event_handler(self, event_id, handler):
        self.evt_handlers[event_id] = handler

    def wait_boot(self):
        try:
            return self.reply(self.REQ_BOOT)
        except:
            # Over USB, reboots cause a reconnect
            self.dev.close()
            print("Waiting for reconnection... ", end="")
            sys.stdout.flush()
            for i in range(200):
                print(".", end="")
                sys.stdout.flush()
                try:
                    self.dev.open()
                except serial.serialutil.SerialException:
                    time.sleep(0.1)
                else:
                    break
            else:
                raise UartTimeout("Reconnection timed out")
            print(" Connected")

    def wait_and_handle_boot(self):
        self.handle_boot(self.wait_boot())

    def nop(self):
        features = Feature.get_all()

        # Send the supported feature flags in the NOP message (has no effect
        # if the target does not support it)
        self.cmd(self.REQ_NOP, struct.pack("<Q", features.value))
        result = self.reply(self.REQ_NOP)

        # Get the enabled feature flags from the message response (returns
        # 0 if the target does not support it)
        features = Feature(struct.unpack("<QQQ", result)[0])

        if self.debug:
            print(f"Enabled features: {features}")

        self.enabled_features = features

    def proxyreq(self, req, reboot=False, no_reply=False, pre_reply=None):
        self.cmd(self.REQ_PROXY, req)
        if pre_reply:
            pre_reply()
        if no_reply:
            return
        elif reboot:
            return self.wait_boot()
        else:
            return self.reply(self.REQ_PROXY)

    def writemem(self, addr, data, progress=False):
        checksum = self.data_checksum(data)
        size = len(data)
        req = struct.pack("<QQI", addr, size, checksum)
        self.cmd(self.REQ_MEMWRITE, req)
        if self.debug:
            print("<< DATA:")
            chexdump(data)
        for i in range(0, len(data), 8192):
            self.dev.write(data[i:i + 8192])
            if progress:
                sys.stdout.write(".")
                sys.stdout.flush()
        if progress:
            print()
        if self.enabled_features & Feature.DISABLE_DATA_CSUMS:
            # Extra sentinel after the data to make sure no data is lost
            self.dev.write(struct.pack("<I", self.DATA_END_SENTINEL))

        # should automatically report a CRC failure
        self.reply(self.REQ_MEMWRITE)

    def readmem(self, addr, size):
        if size == 0:
            return b""

        req = struct.pack("<QQ", addr, size)
        self.cmd(self.REQ_MEMREAD, req)
        reply = self.reply(self.REQ_MEMREAD)
        checksum = struct.unpack("<I",reply[:4])[0]
        data = self.readfull(size)
        if self.debug:
            print(">> DATA:")
            chexdump(data)
        ccsum = self.data_checksum(data)
        if checksum != ccsum:
            raise UartChecksumError("Reply data checksum error: Expected 0x%08x, got 0x%08x"%(checksum, ccsum))

        if self.enabled_features & Feature.DISABLE_DATA_CSUMS:
            # Extra sentinel after the data to make sure no data was lost
            sentinel = struct.unpack("<I", self.readfull(4))[0]
            if sentinel != self.DATA_END_SENTINEL:
                raise UartChecksumError(f"Reply data sentinel error: Expected "
                    f"{self.DATA_END_SENTINEL:#x}, got {sentinel:#x}")

        return data

    def readstruct(self, addr, stype):
        return stype.parse(self.readmem(addr, stype.sizeof()))

class ProxyError(RuntimeError):
    pass

class ProxyReplyError(ProxyError):
    pass

class ProxyRemoteError(ProxyError):
    pass

class ProxyCommandError(ProxyRemoteError):
    pass

class AlignmentError(Exception):
    pass

class IODEV(IntEnum):
    UART = 0
    FB = 1
    USB_VUART = 2
    USB0 = 3
    USB1 = 4
    USB2 = 5
    USB3 = 6
    USB4 = 7
    USB5 = 8
    USB6 = 9
    USB7 = 10

class USAGE(IntFlag):
    CONSOLE = (1 << 0)
    UARTPROXY = (1 << 1)

class GUARD(IntFlag):
    OFF = 0
    SKIP = 1
    MARK = 2
    RETURN = 3
    SILENT = 0x100

REGION_RWX_EL0 = 0x80000000000
REGION_RW_EL0 = 0xa0000000000
REGION_RX_EL1 = 0xc0000000000

# Uses UartInterface.proxyreq() to send requests to M1N1 and process
# reponses sent back.
class M1N1Proxy(Reloadable):
    S_OK = 0
    S_BADCMD = -1

    P_NOP = 0x000
    P_EXIT = 0x001
    P_CALL = 0x002
    P_GET_BOOTARGS = 0x003
    P_GET_BASE = 0x004
    P_SET_BAUD = 0x005
    P_UDELAY = 0x006
    P_SET_EXC_GUARD = 0x007
    P_GET_EXC_COUNT = 0x008
    P_EL0_CALL = 0x009
    P_EL1_CALL = 0x00a
    P_VECTOR = 0x00b
    P_GL1_CALL = 0x00c
    P_GL2_CALL = 0x00d
    P_GET_SIMD_STATE = 0x00e
    P_PUT_SIMD_STATE = 0x00f
    P_REBOOT = 0x010
    P_SLEEP = 0x011
    P_EL3_CALL = 0x012

    P_WRITE64 = 0x100
    P_WRITE32 = 0x101
    P_WRITE16 = 0x102
    P_WRITE8 = 0x103
    P_READ64 = 0x104
    P_READ32 = 0x105
    P_READ16 = 0x106
    P_READ8 = 0x107
    P_SET64 = 0x108
    P_SET32 = 0x109
    P_SET16 = 0x10a
    P_SET8 = 0x10b
    P_CLEAR64 = 0x10c
    P_CLEAR32 = 0x10d
    P_CLEAR16 = 0x10e
    P_CLEAR8 = 0x10f
    P_MASK64 = 0x110
    P_MASK32 = 0x111
    P_MASK16 = 0x112
    P_MASK8 = 0x113
    P_WRITEREAD64 = 0x114
    P_WRITEREAD32 = 0x115
    P_WRITEREAD16 = 0x116
    P_WRITEREAD8 = 0x117

    P_MEMCPY64 = 0x200
    P_MEMCPY32 = 0x201
    P_MEMCPY16 = 0x202
    P_MEMCPY8 = 0x203
    P_MEMSET64 = 0x204
    P_MEMSET32 = 0x205
    P_MEMSET16 = 0x206
    P_MEMSET8 = 0x207

    P_IC_IALLUIS = 0x300
    P_IC_IALLU = 0x301
    P_IC_IVAU = 0x302
    P_DC_IVAC = 0x303
    P_DC_ISW = 0x304
    P_DC_CSW = 0x305
    P_DC_CISW = 0x306
    P_DC_ZVA = 0x307
    P_DC_CVAC = 0x308
    P_DC_CVAU = 0x309
    P_DC_CIVAC = 0x30a
    P_MMU_SHUTDOWN = 0x30b
    P_MMU_INIT = 0x30c
    P_MMU_DISABLE = 0x30d
    P_MMU_RESTORE = 0x30e
    P_MMU_INIT_SECONDARY = 0x30f

    P_XZDEC = 0x400
    P_GZDEC = 0x401

    P_SMP_START_SECONDARIES = 0x500
    P_SMP_CALL = 0x501
    P_SMP_CALL_SYNC = 0x502
    P_SMP_WAIT = 0x503
    P_SMP_SET_WFE_MODE = 0x504
    P_SMP_IS_ALIVE = 0x505
    P_SMP_STOP_SECONDARIES = 0x506
    P_SMP_CALL_EL1 = 0x507
    P_SMP_CALL_SYNC_EL1 = 0x508
    P_SMP_CALL_EL0 = 0x509
    P_SMP_CALL_SYNC_EL0 = 0x50a

    P_HEAPBLOCK_ALLOC = 0x600
    P_MALLOC = 0x601
    P_MEMALIGN = 0x602
    P_FREE = 0x602

    P_KBOOT_BOOT = 0x700
    P_KBOOT_SET_CHOSEN = 0x701
    P_KBOOT_SET_INITRD = 0x702
    P_KBOOT_PREPARE_DT = 0x703

    P_PMGR_CLOCK_ENABLE = 0x800
    P_PMGR_CLOCK_DISABLE = 0x801
    P_PMGR_ADT_CLOCKS_ENABLE = 0x802
    P_PMGR_ADT_CLOCKS_DISABLE = 0x803
    P_PMGR_RESET = 0x804

    P_IODEV_SET_USAGE = 0x900
    P_IODEV_CAN_READ = 0x901
    P_IODEV_CAN_WRITE = 0x902
    P_IODEV_READ = 0x903
    P_IODEV_WRITE = 0x904
    P_IODEV_WHOAMI = 0x905
    P_USB_IODEV_VUART_SETUP = 0x906

    P_TUNABLES_APPLY_GLOBAL = 0xa00
    P_TUNABLES_APPLY_LOCAL = 0xa01

    P_DART_INIT = 0xb00
    P_DART_SHUTDOWN = 0xb01
    P_DART_MAP = 0xb02
    P_DART_UNMAP = 0xb03

    P_HV_INIT = 0xc00
    P_HV_MAP = 0xc01
    P_HV_START = 0xc02
    P_HV_TRANSLATE = 0xc03
    P_HV_PT_WALK = 0xc04
    P_HV_MAP_VUART = 0xc05
    P_HV_TRACE_IRQ = 0xc06
    P_HV_WDT_START = 0xc07
    P_HV_START_SECONDARY = 0xc08
    P_HV_SWITCH_CPU = 0xc09
    P_HV_SET_TIME_STEALING = 0xc0a
    P_HV_PIN_CPU = 0xc0b
    P_HV_WRITE_HCR = 0xc0c
    P_HV_MAP_VIRTIO = 0xc0d
    P_VIRTIO_PUT_BUFFER = 0xc0e
    P_HV_EXIT_CPU = 0xc0f
    P_HV_ADD_TIME = 0xc10

    P_FB_INIT = 0xd00
    P_FB_SHUTDOWN = 0xd01
    P_FB_BLIT = 0xd02
    P_FB_UNBLIT = 0xd03
    P_FB_FILL = 0xd04
    P_FB_CLEAR = 0xd05
    P_FB_DISPLAY_LOGO = 0xd06
    P_FB_RESTORE_LOGO = 0xd07
    P_FB_IMPROVE_LOGO = 0xd08

    P_PCIE_INIT = 0xe00
    P_PCIE_SHUTDOWN = 0xe01

    P_NVME_INIT = 0xf00
    P_NVME_SHUTDOWN = 0xf01
    P_NVME_READ = 0xf02
    P_NVME_FLUSH = 0xf03

    P_MCC_GET_CARVEOUTS = 0x1000

    P_DISPLAY_INIT = 0x1100
    P_DISPLAY_CONFIGURE = 0x1101
    P_DISPLAY_SHUTDOWN = 0x1102
    P_DISPLAY_START_DCP = 0x1103
    P_DISPLAY_IS_EXTERNAL = 0x1104

    P_DAPF_INIT_ALL = 0x1200
    P_DAPF_INIT = 0x1201

    P_CPUFREQ_INIT = 0x1300

    def __init__(self, iface, debug=False):
        self.debug = debug
        self.iface = iface
        self.heap = None

    def _request(self, opcode, *args, reboot=False, signed=False, no_reply=False, pre_reply=None):
        if len(args) > 6:
            raise ValueError("Too many arguments")
        args = list(args) + [0] * (6 - len(args))
        req = struct.pack("<7Q", opcode, *args)
        if self.debug:
            print("<<<< %08x: %08x %08x %08x %08x %08x %08x"%tuple([opcode] + args))
        reply = self.iface.proxyreq(req, reboot=reboot, no_reply=no_reply, pre_reply=None)
        if no_reply or reboot and reply is None:
            return
        ret_fmt = "q" if signed else "Q"
        rop, status, retval = struct.unpack("<Qq" + ret_fmt, reply)
        if self.debug:
            print(">>>> %08x: %d %08x"%(rop, status, retval))
        if reboot:
            return
        if rop != opcode:
            raise ProxyReplyError("Reply opcode mismatch: Expected 0x%08x, got 0x%08x"%(opcode,rop))
        if status != self.S_OK:
            if status == self.S_BADCMD:
                raise ProxyCommandError("Reply error: Bad Command")
            else:
                raise ProxyRemoteError("Reply error: Unknown error (%d)"%status)
        return retval

    def request(self, opcode, *args, **kwargs):
        free = []
        args = list(args)
        args2 = []
        for i, arg in enumerate(args):
            if isinstance(arg, str):
                arg = arg.encode("utf-8") + b"\0"
            if isinstance(arg, bytes) and self.heap:
                p = self.heap.malloc(len(arg))
                free.append(p)
                self.iface.writemem(p, arg)
                if (i < (len(args) - 1)) and args[i + 1] is None:
                    args[i + 1] = len(arg)
                arg = p
            if arg < 0:
                arg &= (1 << 64) - 1
            args2.append(arg)
        try:
            return self._request(opcode, *args2, **kwargs)
        finally:
            for i in free:
                self.heap.free(i)

    def nop(self):
        self.request(self.P_NOP)
    def exit(self, retval=0):
        self.request(self.P_EXIT, retval)
    def call(self, addr, *args, reboot=False):
        if len(args) > 5:
            raise ValueError("Too many arguments")
        return self.request(self.P_CALL, addr, *args, reboot=reboot)
    def reload(self, addr, *args, el1=False):
        if len(args) > 4:
            raise ValueError("Too many arguments")
        if el1:
            self.request(self.P_EL1_CALL, addr, *args, no_reply=True)
        else:
            try:
                self.request(self.P_VECTOR, addr, *args)
                self.iface.wait_boot()
            except ProxyCommandError: # old m1n1 does not support P_VECTOR
                try:
                    self.mmu_shutdown()
                except ProxyCommandError: # older m1n1 does not support MMU
                    pass
                self.request(self.P_CALL, addr, *args, reboot=True)
    def get_bootargs(self):
        return self.request(self.P_GET_BOOTARGS)
    def get_bootargs_rev(self):
        ba_addr = self.request(self.P_GET_BOOTARGS)
        rev = self.read16(ba_addr)
        return (ba_addr, rev)
    def get_base(self):
        return self.request(self.P_GET_BASE)
    def set_baud(self, baudrate):
        self.iface.tty_enable = False
        def change():
            self.iface.dev.baudrate = baudrate
        try:
            self.request(self.P_SET_BAUD, baudrate, 16, 0x005aa5f0, pre_reply=change)
        finally:
            self.iface.tty_enable = True
    def udelay(self, usec):
        self.request(self.P_UDELAY, usec)
    def set_exc_guard(self, mode):
        self.request(self.P_SET_EXC_GUARD, mode)
    def get_exc_count(self):
        return self.request(self.P_GET_EXC_COUNT)
    def el0_call(self, addr, *args):
        if len(args) > 4:
            raise ValueError("Too many arguments")
        return self.request(self.P_EL0_CALL, addr, *args)
    def el1_call(self, addr, *args):
        if len(args) > 4:
            raise ValueError("Too many arguments")
        return self.request(self.P_EL1_CALL, addr, *args)
    def gl1_call(self, addr, *args):
        if len(args) > 4:
            raise ValueError("Too many arguments")
        return self.request(self.P_GL1_CALL, addr, *args)
    def gl2_call(self, addr, *args):
        if len(args) > 4:
            raise ValueError("Too many arguments")
        return self.request(self.P_GL2_CALL, addr, *args)
    def get_simd_state(self, buf):
        self.request(self.P_GET_SIMD_STATE, buf)
    def put_simd_state(self, buf):
        self.request(self.P_PUT_SIMD_STATE, buf)
    def reboot(self):
        self.request(self.P_REBOOT, no_reply=True)
    def sleep(self, deep=False):
        self.request(self.P_SLEEP, deep, no_reply=True)
    def el3_call(self, addr, *args):
        if len(args) > 4:
            raise ValueError("Too many arguments")
        return self.request(self.P_EL3_CALL, addr, *args)

    def write64(self, addr, data):
        '''write 8 byte value to given address'''
        if addr & 7:
            raise AlignmentError()
        self.request(self.P_WRITE64, addr, data)
    def write32(self, addr, data):
        '''write 4 byte value to given address'''
        if addr & 3:
            raise AlignmentError()
        self.request(self.P_WRITE32, addr, data)
    def write16(self, addr, data):
        '''write 2 byte value to given address'''
        if addr & 1:
            raise AlignmentError()
        self.request(self.P_WRITE16, addr, data)
    def write8(self, addr, data):
        '''write 1 byte value to given address'''
        self.request(self.P_WRITE8, addr, data)

    def read64(self, addr):
        '''return 8 byte value from given address'''
        if addr & 7:
            raise AlignmentError()
        return self.request(self.P_READ64, addr)
    def read32(self, addr):
        '''return 4 byte value given address'''
        if addr & 3:
            raise AlignmentError()
        return self.request(self.P_READ32, addr)
    def read16(self, addr):
        '''return 2 byte value from given address'''
        if addr & 1:
            raise AlignmentError()
        return self.request(self.P_READ16, addr)
    def read8(self, addr):
        '''return 1 byte value from given address'''
        return self.request(self.P_READ8, addr)

    def set64(self, addr, data):
        '''Or 64 bit value of data into memory at addr and return result'''
        if addr & 7:
            raise AlignmentError()
        return self.request(self.P_SET64, addr, data)
    def set32(self, addr, data):
        '''Or 32 bit value of data into memory at addr and return result'''
        if addr & 3:
            raise AlignmentError()
        return self.request(self.P_SET32, addr, data)
    def set16(self, addr, data):
        '''Or 16 bit value of data into memory at addr and return result'''
        if addr & 1:
            raise AlignmentError()
        return self.request(self.P_SET16, addr, data)
    def set8(self, addr, data):
        '''Or byte value of data into memory at addr and return result'''
        return self.request(self.P_SET8, addr, data)

    def clear64(self, addr, data):
        '''Clear bits in 64 bit memory at address addr that are set
    in parameter data and return result'''
        if addr & 7:
            raise AlignmentError()
        return self.request(self.P_CLEAR64, addr, data)
    def clear32(self, addr, data):
        '''Clear bits in 32 bit memory at address addr that are set
    in parameter data and return result'''
        if addr & 3:
            raise AlignmentError()
        return self.request(self.P_CLEAR32, addr, data)
    def clear16(self, addr, data):
        '''Clear bits in 16 bit memory at address addr that are set
    in parameter data and return result'''
        if addr & 1:
            raise AlignmentError()
        return self.request(self.P_CLEAR16, addr, data)
    def clear8(self, addr, data):
        '''Clear bits in 8 bit memory at addr that are set in data
    and return result'''
        return self.request(self.P_CLEAR8, addr, data)

    def mask64(self, addr, clear, set):
        '''Clear bits in 64 bit memory at address addr that are
 set in clear, then set the bits in set and return result'''
        if addr & 7:
            raise AlignmentError()
        return self.request(self.P_MASK64, addr, clear, set)
    def mask32(self, addr, clear, set):
        '''Clear bits in 32 bit memory at address addr that are
 set in clear, then set the bits in set and return result'''
        if addr & 3:
            raise AlignmentError()
        return self.request(self.P_MASK32, addr, clear, set)
    def mask16(self, addr, clear, set):
        '''Clear select bits in 16 bit memory addr that are set
 in clear parameter, then set the bits in set parameter and return result'''
        if addr & 1:
            raise AlignmentError()
        return self.request(self.P_MASK16, addr, clear, set)
    def mask8(self, addr, clear, set):
        '''Clear bits in 1 byte memory at addr that are set
 in clear parameter, then set the bits in set parameter
 and return the result'''
        return self.request(self.P_MASK8, addr, clear, set)

    def writeread64(self, addr, data):
        return self.request(self.P_WRITEREAD64, addr, data)
    def writeread32(self, addr, data):
        return self.request(self.P_WRITEREAD32, addr, data)
    def writeread16(self, addr, data):
        return self.request(self.P_WRITEREAD16, addr, data)
    def writeread8(self, addr, data):
        return self.request(self.P_WRITEREAD8, addr, data)

    def memcpy64(self, dst, src, size):
        if src & 7 or dst & 7:
            raise AlignmentError()
        self.request(self.P_MEMCPY64, dst, src, size)
    def memcpy32(self, dst, src, size):
        if src & 3 or dst & 3:
            raise AlignmentError()
        self.request(self.P_MEMCPY32, dst, src, size)
    def memcpy16(self, dst, src, size):
        if src & 1 or dst & 1:
            raise AlignmentError()
        self.request(self.P_MEMCPY16, dst, src, size)
    def memcpy8(self, dst, src, size):
        self.request(self.P_MEMCPY8, dst, src, size)

    def memset64(self, dst, src, size):
        if dst & 7:
            raise AlignmentError()
        self.request(self.P_MEMSET64, dst, src, size)
    def memset32(self, dst, src, size):
        if dst & 3:
            raise AlignmentError()
        self.request(self.P_MEMSET32, dst, src, size)
    def memset16(self, dst, src, size):
        if dst & 1:
            raise AlignmentError()
        self.request(self.P_MEMSET16, dst, src, size)
    def memset8(self, dst, src, size):
        self.request(self.P_MEMSET8, dst, src, size)

    def ic_ialluis(self):
        self.request(self.P_IC_IALLUIS)
    def ic_iallu(self):
        self.request(self.P_IC_IALLU)
    def ic_ivau(self, addr, size):
        self.request(self.P_IC_IVAU, addr, size)
    def dc_ivac(self, addr, size):
        self.request(self.P_DC_IVAC, addr, size)
    def dc_isw(self, sw):
        self.request(self.P_DC_ISW, sw)
    def dc_csw(self, sw):
        self.request(self.P_DC_CSW, sw)
    def dc_cisw(self, sw):
        self.request(self.P_DC_CISW, sw)
    def dc_zva(self, addr, size):
        self.request(self.P_DC_ZVA, addr, size)
    def dc_cvac(self, addr, size):
        self.request(self.P_DC_CVAC, addr, size)
    def dc_cvau(self, addr, size):
        self.request(self.P_DC_CVAU, addr, size)
    def dc_civac(self, addr, size):
        self.request(self.P_DC_CIVAC, addr, size)
    def mmu_shutdown(self):
        self.request(self.P_MMU_SHUTDOWN)
    def mmu_init(self):
        self.request(self.P_MMU_INIT)
    def mmu_disable(self):
        return self.request(self.P_MMU_DISABLE)
    def mmu_restore(self, flags):
        self.request(self.P_MMU_RESTORE, flags)
    def mmu_init_secondary(self, cpu):
        self.request(self.P_MMU_INIT_SECONDARY, cpu)


    def xzdec(self, inbuf, insize, outbuf=0, outsize=0):
        return self.request(self.P_XZDEC, inbuf, insize, outbuf,
                            outsize, signed=True)

    def gzdec(self, inbuf, insize, outbuf, outsize):
        return self.request(self.P_GZDEC, inbuf, insize, outbuf,
                            outsize, signed=True)

    def smp_start_secondaries(self):
        self.request(self.P_SMP_START_SECONDARIES)
    def smp_call(self, cpu, addr, *args):
        if len(args) > 4:
            raise ValueError("Too many arguments")
        self.request(self.P_SMP_CALL, cpu, addr, *args)
    def smp_call_sync(self, cpu, addr, *args):
        if len(args) > 4:
            raise ValueError("Too many arguments")
        return self.request(self.P_SMP_CALL_SYNC, cpu, addr, *args)
    def smp_wait(self, cpu):
        return self.request(self.P_SMP_WAIT, cpu)
    def smp_set_wfe_mode(self, mode):
        return self.request(self.P_SMP_SET_WFE_MODE, mode)
    def smp_is_alive(self, cpu):
        return self.request(self.P_SMP_IS_ALIVE, cpu)
    def smp_stop_secondaries(self, deep_sleep=False):
        self.request(self.P_SMP_STOP_SECONDARIES, deep_sleep)
    def smp_call_el1(self, cpu, addr, *args):
        if len(args) > 3:
            raise ValueError("Too many arguments")
        self.request(self.P_SMP_CALL_EL1, cpu, addr, *args)
    def smp_call_sync_el1(self, cpu, addr, *args):
        if len(args) > 3:
            raise ValueError("Too many arguments")
        return self.request(self.P_SMP_CALL_SYNC_EL1, cpu, addr, *args)
    def smp_call_el0(self, cpu, addr, *args):
        if len(args) > 3:
            raise ValueError("Too many arguments")
        self.request(self.P_SMP_CALL_EL0, cpu, addr, *args)
    def smp_call_sync_el0(self, cpu, addr, *args):
        if len(args) > 3:
            raise ValueError("Too many arguments")
        return self.request(self.P_SMP_CALL_SYNC_EL0, cpu, addr, *args)

    def heapblock_alloc(self, size):
        return self.request(self.P_HEAPBLOCK_ALLOC, size)
    def malloc(self, size):
        return self.request(self.P_MALLOC, size)
    def memalign(self, align, size):
        return self.request(self.P_MEMALIGN, align, size)
    def free(self, ptr):
        self.request(self.P_FREE, ptr)

    def kboot_boot(self, kernel):
        self.request(self.P_KBOOT_BOOT, kernel)
    def kboot_set_chosen(self, name, value):
        self.request(self.P_KBOOT_SET_CHOSEN, name, value)
    def kboot_set_initrd(self, base, size):
        self.request(self.P_KBOOT_SET_INITRD, base, size)
    def kboot_prepare_dt(self, dt_addr):
        return self.request(self.P_KBOOT_PREPARE_DT, dt_addr)

    def pmgr_clock_enable(self, clkid):
        return self.request(self.P_PMGR_CLOCK_ENABLE, clkid)
    def pmgr_clock_disable(self, clkid):
        return self.request(self.P_PMGR_CLOCK_DISABLE, clkid)
    def pmgr_adt_clocks_enable(self, path):
        return self.request(self.P_PMGR_ADT_CLOCKS_ENABLE, path)
    def pmgr_adt_clocks_disable(self, path):
        return self.request(self.P_PMGR_ADT_CLOCKS_DISABLE, path)
    def pmgr_reset(self, die, name):
        return self.request(self.P_PMGR_RESET, die, name)

    def iodev_set_usage(self, iodev, usage):
        return self.request(self.P_IODEV_SET_USAGE, iodev, usage)
    def iodev_can_read(self, iodev):
        return self.request(self.P_IODEV_CAN_READ, iodev)
    def iodev_can_write(self, iodev):
        return self.request(self.P_IODEV_CAN_WRITE, iodev)
    def iodev_read(self, iodev, buf, size=None):
        return self.request(self.P_IODEV_READ, iodev, buf, size)
    def iodev_write(self, iodev, buf, size=None):
        return self.request(self.P_IODEV_WRITE, iodev, buf, size)
    def iodev_whoami(self):
        return IODEV(self.request(self.P_IODEV_WHOAMI))
    def usb_iodev_vuart_setup(self, iodev):
        return self.request(self.P_USB_IODEV_VUART_SETUP, iodev)

    def tunables_apply_global(self, path, prop):
        return self.request(self.P_TUNABLES_APPLY_GLOBAL, path, prop)
    def tunables_apply_local(self, path, prop, reg_offset):
        return self.request(self.P_TUNABLES_APPLY_LOCAL, path, prop, reg_offset)
    def tunables_apply_local_addr(self, path, prop, base):
        return self.request(self.P_TUNABLES_APPLY_LOCAL, path, prop, base)

    def dart_init(self, base, sid, dart_type=DART.T8020):
        return self.request(self.P_DART_INIT, base, sid, dart_type)
    def dart_shutdown(self, dart):
        return self.request(self.P_DART_SHUTDOWN, dart)
    def dart_map(self, dart, iova, bfr, len):
        return self.request(self.P_DART_MAP, dart, iova, bfr, len)
    def dart_unmap(self, dart, iova, len):
        return self.request(self.P_DART_UNMAP, dart, iova, len)

    def hv_init(self):
        return self.request(self.P_HV_INIT)
    def hv_map(self, from_, to, size, incr):
        return self.request(self.P_HV_MAP, from_, to, size, incr)
    def hv_start(self, entry, *args):
        return self.request(self.P_HV_START, entry, *args)
    def hv_translate(self, addr, s1=False, w=False):
        '''Translate virtual address
 stage 1 only if s1, for write if w'''
        return self.request(self.P_HV_TRANSLATE, addr, s1, w)
    def hv_pt_walk(self, addr):
        return self.request(self.P_HV_PT_WALK, addr)
    def hv_map_vuart(self, base, irq, iodev):
        return self.request(self.P_HV_MAP_VUART, base, irq, iodev)
    def hv_trace_irq(self, evt_type, num, count, flags):
        return self.request(self.P_HV_TRACE_IRQ, evt_type, num, count, flags)
    def hv_wdt_start(self, cpu):
        return self.request(self.P_HV_WDT_START, cpu)
    def hv_start_secondary(self, cpu, entry, *args):
        return self.request(self.P_HV_START_SECONDARY, cpu, entry, *args)
    def hv_switch_cpu(self, cpu):
        return self.request(self.P_HV_SWITCH_CPU, cpu)
    def hv_set_time_stealing(self, enabled, reset):
        return self.request(self.P_HV_SET_TIME_STEALING, int(bool(enabled)), int(bool(reset)))
    def hv_pin_cpu(self, cpu):
        return self.request(self.P_HV_PIN_CPU, cpu)
    def hv_write_hcr(self, hcr):
        return self.request(self.P_HV_WRITE_HCR, hcr)
    def hv_map_virtio(self, base, config):
        return self.request(self.P_HV_MAP_VIRTIO, base, config)
    def virtio_put_buffer(self, base, qu, idx, length):
        return self.request(self.P_VIRTIO_PUT_BUFFER, base, qu, idx, length)
    def hv_exit_cpu(self, cpu=-1):
        return self.request(self.P_HV_EXIT_CPU, cpu)
    def hv_add_time(self, time):
        return self.request(self.P_HV_ADD_TIME, time)

    def fb_init(self):
        return self.request(self.P_FB_INIT)
    def fb_shutdown(self, restore_logo=True):
        return self.request(self.P_FB_SHUTDOWN, restore_logo)
    def fb_blit(self, x, y, w, h, ptr, stride, pix_fmt=PIX_FMT.XRGB):
        return self.request(self.P_FB_BLIT, x, y, w, h, ptr, stride | pix_fmt << 32)
    def fb_unblit(self, x, y, w, h, ptr, stride):
        return self.request(self.P_FB_UNBLIT, x, y, w, h, ptr, stride)
    def fb_fill(self, x, y, w, h, color):
        return self.request(self.P_FB_FILL, x, y, w, h, color)
    def fb_clear(self, color):
        return self.request(self.P_FB_CLEAR, color)
    def fb_display_logo(self):
        return self.request(self.P_FB_DISPLAY_LOGO)
    def fb_restore_logo(self):
        return self.request(self.P_FB_RESTORE_LOGO)
    def fb_improve_logo(self):
        return self.request(self.P_FB_IMPROVE_LOGO)

    def pcie_init(self):
        return self.request(self.P_PCIE_INIT)
    def pcie_shutdown(self):
        return self.request(self.P_PCIE_SHUTDOWN)

    def nvme_init(self):
        return self.request(self.P_NVME_INIT)
    def nvme_shutdown(self):
        return self.request(self.P_NVME_SHUTDOWN)
    def nvme_read(self, nsid, lba, bfr):
        return self.request(self.P_NVME_READ, nsid, lba, bfr)
    def nvme_flush(self, nsid):
        return self.request(self.P_NVME_FLUSH, nsid)

    def mcc_get_carveouts(self):
        return self.request(self.P_MCC_GET_CARVEOUTS)

    def display_init(self):
        return self.request(self.P_DISPLAY_INIT)
    def display_configure(self, cfg):
        return self.request(self.P_DISPLAY_CONFIGURE, cfg)
    def display_shutdown(self, mode):
        return self.request(self.P_DISPLAY_SHUTDOWN, mode)
    def display_start_dcp(self):
        return self.request(self.P_DISPLAY_START_DCP)
    def display_is_external(self):
        return self.request(self.P_DISPLAY_IS_EXTERNAL)

    def dapf_init_all(self):
        return self.request(self.P_DAPF_INIT_ALL)
    def dapf_init(self, path):
        return self.request(self.P_DAPF_INIT, path)

    def cpufreq_init(self):
        return self.request(self.P_CPUFREQ_INIT)

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)

# To run m1n1/proxy.py as script, use:
# $ cd <path to>/m1n1/proxyclient
# $ python3 -m m1n1.proxy m1n1/proxy.py
if __name__ == "__main__":
    uartif = UartInterface(device=None, debug=True)
    print("Sending NOP...", end=' ')
    uartif.nop()
    print("OK")
    proxy = M1N1Proxy(uartif, debug=True)
    print("Sending Proxy NOP...", end=' ')
    proxy.nop()
    print("OK")
    print("Boot args: 0x%x" % proxy.get_bootargs())



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, argparse
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))
sys.path.append("/home/eileen/asahi/avd")  # git clone https://github.com/eiln/avd.git
# Decode via AVD Cortex-M3 firmware

from m1n1.setup import *
from m1n1.utils import *
from m1n1.fw.avd import *
import cv2

from avd_emu import AVDEmulator
from avid.h264.decoder import AVDH264Decoder
from avid.vp9.decoder import AVDVP9Decoder
from avid.utils import *
from tools.common import *

if __name__ == "__main__":
	parser = argparse.ArgumentParser()
	parser.add_argument('-f','--firmware', type=str, default="data/fw.bin", help="path to CM3 firmware")
	parser.add_argument('-i','--input', type=str, required=True, help="path to CM3 firmware")
	parser.add_argument('-d','--dir', type=str, required=True, help="frame_params trace dir")
	parser.add_argument('-p', '--poll', action='store_true', help="poll iommu space")
	args = parser.parse_args()

	paths = os.listdir(os.path.join(args.dir))
	paths = sorted([os.path.join(args.dir, path) for path in paths if "frame" in path])
	assert(len(paths))

	avd = AVDDevice(u)
	avd.decoder = AVDVP9Dec(avd)
	avd.boot()
	avd.mcpu_decode_init(args.firmware)
	avd.poll()

	avd.ioalloc_at(0x0, 0xff0000, stream=0)
	if (args.poll):
		avd.iomon.add(0x0, 0xff0000)
	avd.ioalloc_at(0x0, 0xb84000, stream=1)
	avd.iomon.poll()
	emu = AVDEmulator(args.firmware, stfu=True)
	emu.start()

	dec = AVDVP9Decoder()
	dec.stfu = True
	dec.hal.stfu = True
	units = dec.setup(args.input, num=4, do_probs=True)
	for n,unit in enumerate(units):
		inst = dec.decode(unit)
		avd.decoder.set_payload(dec.ctx, units[n])
		avd.decoder.avd.iowrite(dec.ctx.probs_addr, unit.get_probs(), stream=0)
		avd.iomon.poll()
		cmd = emu.set_params(paths[n])
		xxde(cmd)
		avd.iowrite(0x0, emu.dart1_space, stream=1)
		avd.iomon.poll()
		avd.avd_wbuf(emu.get_cmd_addr(n), cmd)
		avd.avd_w32(0x1098054, emu.get_cmd_addr(n))
		avd.poll()
		avd.iomon.poll()
		img = avd.decoder.get_nv12_disp_frame(dec.ctx)
		cv2.imshow(avd.decoder.winname, img); cv2.waitKey(1)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct
from construct import *

from m1n1.setup import *
from m1n1.shell import run_shell
from m1n1 import asm
from m1n1.hw.dart import DART
from m1n1.fw.dcp.client import DCPClient
from m1n1.fw.dcp.manager import DCPManager
from m1n1.fw.dcp.ipc import ByRef
from m1n1.proxyutils import RegMonitor

disp_name = "/arm-io/disp0"

external = hasattr(u.adt[disp_name], "external") and u.adt[disp_name].external != 0
compat = u.adt[disp_name].compatible[0].split(",")[-1]

mon = RegMonitor(u)

if compat == 't8103':
    #mon.add(0x230000000, 0x18000)
    #mon.add(0x230018000, 0x4000)
    #mon.add(0x230068000, 0x8000)
    #mon.add(0x2300b0000, 0x8000)
    #mon.add(0x2300f0000, 0x4000)
    #mon.add(0x230100000, 0x10000)
    #mon.add(0x230170000, 0x10000)
    #mon.add(0x230180000, 0x1c000)
    #mon.add(0x2301a0000, 0x10000)
    #mon.add(0x2301d0000, 0x4000)
    #mon.add(0x230230000, 0x10000)
    #mon.add(0x23038c000, 0x10000)
    #mon.add(0x230800000, 0x10000)
    #mon.add(0x230840000, 0xc000)
    #mon.add(0x230850000, 0x2000)
    ##mon.add(0x230852000, 0x5000) # big curve / gamma table
    #mon.add(0x230858000, 0x18000)
    #mon.add(0x230870000, 0x4000)
    #mon.add(0x230880000, 0x8000)
    #mon.add(0x230894000, 0x4000)
    #mon.add(0x2308a8000, 0x8000)
    #mon.add(0x2308b0000, 0x8000)
    #mon.add(0x2308f0000, 0x4000)
    ##mon.add(0x2308fc000, 0x4000) # stats / RGB color histogram
    #mon.add(0x230900000, 0x10000)
    #mon.add(0x230970000, 0x10000)
    #mon.add(0x230980000, 0x10000)
    #mon.add(0x2309a0000, 0x10000)
    #mon.add(0x2309d0000, 0x4000)
    #mon.add(0x230a30000, 0x20000)
    #mon.add(0x230b8c000, 0x10000)
    #mon.add(0x231100000, 0x8000)
    #mon.add(0x231180000, 0x4000)
    #mon.add(0x2311bc000, 0x10000)
    #mon.add(0x231300000, 0x8000)
    ##mon.add(0x23130c000, 0x4000) # - DCP dart
    #mon.add(0x231310000, 0x8000)
    #mon.add(0x231340000, 0x8000)
    ##mon.add(0x231800000, 0x8000) # breaks DCP
    ##mon.add(0x231840000, 0x8000) # breaks DCP
    ##mon.add(0x231850000, 0x8000) # something DCP?
    ##mon.add(0x231920000, 0x8000) # breaks DCP
    ##mon.add(0x231960000, 0x8000) # breaks DCP
    ##mon.add(0x231970000, 0x10000) # breaks DCP
    ##mon.add(0x231c00000, 0x10000) # DCP mailbox

    mon.add(0x230845840, 0x40) # error regs

def get_color_mode(mgr):
    best_id = None
    best_score = -1
    for mode in mgr.dcpav_prop['ColorElements']:
        if mode['IsVirtual']:
            continue
        if mode['Depth'] != 8:
            continue
        if mode['Score'] > best_score:
            best_score = mode['Score']
            best_id =  mode['ID']
    return best_id

def get_timing_mode(mgr):
    best_id = None
    best_score = -1
    for mode in mgr.dcpav_prop['TimingElements']:
        if mode['IsVirtual']:
            continue
        if int(mode['Score']) > best_score:
            best_score = int(mode['Score'])
            best_id =  int(mode['ID'])
    return best_id

mon.poll()

dart = DART.from_adt(u, "arm-io/dart-dcp")
disp_dart = DART.from_adt(u, "arm-io/dart-disp0")

print("DCP DART:")
dart.dart.regs.dump_regs()
print("DISP DART:")
disp_dart.dart.regs.dump_regs()

dcp_addr = u.adt["arm-io/dcp"].get_reg(0)[0]
dcp = DCPClient(u, dcp_addr, dart, disp_dart)
dcp.dva_offset = getattr(u.adt["/arm-io/dcp"][0], "asc_dram_mask", 0)

dcp.start()
dcp.start_ep(0x20)
dcp.start_ep(0x37)
dcp.dcpep.initialize()

dcp.system.wait_for("system")
dcp.system.system.setProperty("gAFKConfigLogMask", 0xffff)

mgr = DCPManager(dcp.dcpep, compat)

mon.poll()

mgr.start_signal()

mon.poll()

mgr.get_color_remap_mode(6)
mgr.enable_disable_video_power_savings(0)

mgr.update_notify_clients_dcp([0,0,0,0,0,0,1,1,1,0,1,1,1,1])
mgr.first_client_open()
print(f"keep on: {mgr.isKeepOnScreen()}")
print(f"main display: {mgr.is_main_display()}")
assert mgr.setPowerState(1, False, ByRef(0)) == 0

mon.poll()

if external:
    assert mgr.set_display_device(2) == 0
else:
    assert mgr.set_display_device(0) == 2
assert mgr.set_parameter_dcp(14, [0], 1) == 0

color_mode = get_color_mode(mgr)
timing_mode = get_timing_mode(mgr)
mgr.SetDigitalOutMode(color_mode, timing_mode)
mon.poll()

while mgr.iomfb_prop['DPTimingModeId'] != timing_mode:
    print("Try re-setting mode")
    mgr.SetDigitalOutMode(color_mode, timing_mode)
    mon.poll()

if external:
    assert mgr.set_display_device(2) == 0
else:
    assert mgr.set_display_device(0) == 2
assert mgr.set_parameter_dcp(14, [0], 1) == 0

t = ByRef(b"\x00" * 0xc0c)
assert mgr.get_gamma_table(t) == 2
assert mgr.set_contrast(0) == 0
assert mgr.setBrightnessCorrection(65536) == 0

if external:
    assert mgr.set_display_device(2) == 0
else:
    assert mgr.set_display_device(0) == 2
assert mgr.set_parameter_dcp(14, [0], 1) == 0

mon.poll()

swapid = ByRef(0)

def start():
    # arg: IOUserClient
    ret = mgr.swap_start(swapid, {
        "addr": 0xFFFFFE1667BA4A00,
        "unk": 0,
        "flag1": 0,
        "flag2": 1
    })
    assert ret == 0
    print(f"swap ID: {swapid.val:#x}")

start()

mgr.set_matrix(9, [[1<<32, 0, 0],
                   [0, 1<<32, 0],
                   [0, 0, 1<<32]])
mgr.setBrightnessCorrection(65536)
mgr.set_parameter_dcp(3, [65536], 1)
mgr.set_parameter_dcp(6, [65536], 1)

width = mgr.display_width()
height = mgr.display_height()

surface_id = 3

swap_rec = Container(
    flags1 = 0x861202,
    flags2 = 0x04,
    swap_id = swapid.val,
    surf_ids = [surface_id, 0, 0, 0],
    src_rect = [[0, 0, width, height],[0,0,0,0],[0,0,0,0],[0,0,0,0]],
    surf_flags = [1, 0, 0, 0],
    surf_unk = [0, 0, 0, 0],
    dst_rect = [[0, 0, width, height],[0,0,0,0],[0,0,0,0],[0,0,0,0]],
    swap_enabled = 0x80000007,
    swap_completed = 0x80000007,
    bl_unk = 0x1,
    bl_val = 0x58f058d0, # ~99 nits
    bl_power = 0x40,
)

surf = Container(
    is_tiled = False,
    unk_1 = False,
    unk_2 = False,
    plane_cnt = 0,
    plane_cnt2 = 0,
    format = "BGRA",
    xfer_func = 13,
    colorspace = 1,
    stride = width * 4,
    pix_size = 4,
    pel_w = 1,
    pel_h = 1,
    offset = 0,
    width = width,
    height = height,
    buf_size = width * height * 4,
    surface_id = surface_id,
    has_comp = True,
    has_planes = True,
    has_compr_info = False,
    unk_1f5 = 0,
    unk_1f9 = 0,
)

compressed_surf = Container(
    is_tiled = False,
    unk_1 = False,
    unk_2 = False,
    plane_cnt = 2,
    plane_cnt2 = 2,
    format = 'b3a8',
    unk_f = 0x00000000,
    xfer_func = 13,
    colorspace = 2,
    stride = width,
    pix_size = 1,
    pel_w = 1,
    pel_h = 1,
    offset = 0,
    width = width,
    height = height,
    buf_size = 0x00A36000,
    unk_2d = 0,
    unk_31 = 0,
    surface_id = 5,
    comp_types = [
        Container(count = 0, types =[]),
        Container(count = 0, types =[]),
    ],
    has_comp = True,
    planes = [
        Container(
            width = width,
            height = height,
            base = 0,
            offset = 0,
            stride = 0x1e000,
            size = 0x818000,
            tile_size = 1024,
            tile_w = 16,
            tile_h = 16,
            unk2 = 0x05,
        ),
        Container(
            width = width,
            height = height,
            base = 0x818000,
            offset = 0x818000,
            stride = 0x7800,
            size = 0x21e000,
            tile_size = 256,
            tile_w = 16,
            tile_h = 16,
            unk2 = 0x05,
        )
    ],
    has_planes = True,
    compression_info = [
        unhex("""
            10 00 00 00 10 00 00 00 00 80 7F 00 00 00 00 00
            08 00 00 00 78 00 00 00 44 00 00 00 00 00 00 00
            03 00 00 00 00 00 00 00 AA AA AA 00 04 00 00 00
            E0 01 00 AA
        """),
        unhex("""
            10 00 00 00 10 00 00 00 00 60 A1 00 00 80 81 00
            08 00 00 00 78 00 00 00 44 00 00 00 00 00 00 00
            03 00 00 00 00 00 00 00 AA AA AA 00 01 00 00 00
            78 00 00 AA
        """),
    ],
    has_compr_info = True,
    unk_1f5 = 0x100000,
    unk_1f9 = 0x100000,
)


outB = ByRef(False)

swaps = mgr.swaps

mon.poll()

fb_size = align_up(width * height * 4, 8 * 0x4000)
print(f"Display {width}x{height}, fb size: {fb_size}")

buf = u.memalign(0x4000, fb_size)

colors = [0xDD0000, 0xFE6230, 0xFEF600, 0x00BB00, 0x009BFE, 0x000083, 0x30009B]


for i, color in enumerate(colors):
    lines = height // len(colors)
    offset = i * lines * width * 4
    p.memset32(buf + offset, color, lines * width * 4)

iova = disp_dart.iomap(0, buf, fb_size)

surfaces = [surf, None, None, None]
#surfaces = [compressed_surf, None, None, None]
surfAddr = [iova, 0, 0, 0]

def submit():
    swap_rec.swap_id = swapid.val
    ret = mgr.swap_submit_dcp(swap_rec=swap_rec, surfaces=surfaces, surfAddr=surfAddr,
                            unkBool=False, unkFloat=0.0, unkInt=0, unkOutBool=outB)
    print(f"swap returned {ret} / {outB}")

    dcp.work()

    if ret == 0:
        while swaps == mgr.swaps:
            dcp.work()
        print("swap complete!")

submit()

run_shell(globals(), msg="Have fun!")



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.loadobjs import *
import argparse
import numpy as np

argparser = argparse.ArgumentParser()
argparser.add_argument("-d", "--domain", type=str,
		       help='Look for MMIO range associated with a particular'
		       		' power domain')
argparser.add_argument("-p", "--print", action='store_true',
		       help='Print power domain list')
args = argparser.parse_args()

if args.print:
	for dev in u.adt["/arm-io/pmgr"].devices:
		print(dev.name)
	sys.exit(0)

granule = 0x4000

lp = LinkedProgram(u)
lp.load_inline_c(f'''
	#define GRANULE {granule}
	''' + '''
	#include "exception.h"
	#include "utils.h"
	#include "soc.h"

	bool is_t6000(void)
	{
		return chip_id == T6000;
	}

	void sweep(u64 from, u64 to, u64 target)
	{
		u32 *mask = (u32 *) target;
		exc_guard = GUARD_MARK | GUARD_SILENT;

		int bitp = 0;
		for (u64 p = from; p < to; p += GRANULE) {
			sysop("dsb sy");
			sysop("isb");
			bool hit = read32(p) != 0xabad1dea;

			if (hit)
				*mask |= (1 << bitp);
			else
				*mask &= ~(1 << bitp);

			if (++bitp >= 32) {
				bitp = 0;
				mask++;
			}
		}

		sysop("dsb sy");
		sysop("isb");
	}
	''')

def do_sweep(maskrange):
	masklen = (maskrange.stop - maskrange.start) // granule // 32 * 4 + 4
	mask_base = u.heap.malloc(masklen)
	lp.sweep(maskrange.start, maskrange.stop, mask_base)
	mask = iface.readmem(mask_base, masklen)
	u.heap.free(mask_base)
	return np.frombuffer(mask, dtype=np.uint8)

def describe_mask(mask, maskrange):
	'''
	Describe mask in terms of hot from-to ranges
	'''
	ranges = []
	prev_hit = False
	mask = np.concatenate((mask, [0]))
	for i in range(len(mask)*8):
		hit = mask[i//8] & (1<<(i%8)) != 0
		if hit and not prev_hit:
			start = maskrange.start + i*granule
		if not hit and prev_hit:
			end = maskrange.start + i*granule
			ranges.append((start, end))
		prev_hit = hit
	return ranges

if lp.is_t6000():
	maskrange = range(0x2_9000_0000, 0x4_0000_0000)
else:
	maskrange = range(0x2_2000_0000, 0x3_0000_0000)

pd_did_enable = set()
pmgr = u.adt["/arm-io/pmgr"]
ps_dev_by_id = {dev.id: dev for dev in pmgr.devices}
ps_deps = dict()
ps_addrs = dict()

for dev in pmgr.devices:
	ps = pmgr.ps_regs[dev.psreg]
	addr = pmgr.get_reg(ps.reg)[0] + ps.offset + dev.psidx * 8

	if lp.is_t6000() and dev.name.startswith("AOP_"):
		addr = 0x292284000 + (dev.id - 403) * 8		

	ps_addrs[dev.name] = addr
	ps_deps[dev.name] = [
		ps_dev_by_id[idx].name for idx
		in dev.parents if idx in ps_dev_by_id
	]

if lp.is_t6000():
	# on t6000, guess the AOP PD hierarchy (undocumented
	# in ADT) by analogy with t8103
	ps_deps["AOP_GPIO"] += ["AOP_FILTER"]
	ps_deps["AOP_BASE"] += ["AOP_FILTER"]
	ps_deps["AOP_FR"] += ["AOP_FILTER"]
	ps_deps["AOP_SPMI0"] += ["AOP_FR"]
	ps_deps["AOP_SPMI1"] += ["AOP_FR"]
	ps_deps["AOP_LEAP_CLK"] += ["AOP_FILTER"]
	ps_deps["AOP_SHIM"] += ["AOP_BASE"]
	ps_deps["AOP_UART0"] += ["AOP_SHIM"]
	ps_deps["AOP_UART1"] += ["AOP_SHIM"]
	ps_deps["AOP_UART2"] += ["AOP_SHIM"]
	ps_deps["AOP_SCM"] += ["AOP_BASE", "AOP_FR"]
	ps_deps["AOP_CPU"] += ["AOP_BASE"]
	ps_deps["AOP_I2CM0"] += ["AOP_FR"]
	ps_deps["AOP_I2CM1"] += ["AOP_FR"]
	ps_deps["AOP_MCA0"] += ["AOP_FR", "AOP_SHIM"]
	ps_deps["AOP_MCA1"] += ["AOP_FR", "AOP_SHIM"]
	ps_deps["AOP_SPI0"] += ["AOP_FR"]
	ps_deps["AOP_LEAP"] += ["AOP_LEAP_CLK"]
	ps_deps["AOP_AUDIO_SHIM"] += ["AOP_LEAP_CLK"]
	ps_deps["AOP_AUDIO_ADMA0"] += ["AOP_FR"]
	ps_deps["AOP_PDMC_LPD"] += ["AOP_SHIM"]
	ps_deps["AOP_SRAM"] += ["AOP_SCM", "AOP_CPU"]

def ps_pstate(name):
	return p.read32(ps_addrs[name]) & 0x0f

def ps_enabled(name):
	return p.read32(ps_addrs[name]) & 0x0f == 0x0f

def ps_set_pstate(name, desired):
	p.mask32(ps_addrs[name], 0xf, desired)
	time.sleep(0.001)
	actual = p.read32(ps_addrs[name])
	if actual & 0xf0 != desired << 4:
		print("WARNING: %s stuck at pstate 0x%x (desired 0x%x)" \
			% (name, actual >> 4, desired))

def ps_enable(name):
	print("Enabling %s..." % name)
	ps_set_pstate(name, 0xf)

def ps_disable(name):
	p.mask32(ps_addrs[name], 0xf, 0x0)

if args.domain:
	ps_disable(args.domain)

	to_enable = set([args.domain])
	for dev in reversed(pmgr.devices):
		if dev.name not in to_enable \
				or ps_enabled(dev.name):
			continue

		for dep in ps_deps[dev.name]:
			to_enable.add(dep)

	save = dict()
	for dev in pmgr.devices:
		if dev.name in to_enable:
			save[dev.name] = ps_pstate(dev.name)
			if dev.name != args.domain:
				ps_enable(dev.name)

	premask = do_sweep(maskrange)
	ps_enable(args.domain)
	postmask = do_sweep(maskrange)

	print("Reverting...")

	for dev in reversed(pmgr.devices):
		if dev.name in to_enable and dev.name:
			ps_set_pstate(dev.name, save[dev.name])

	hitmask = premask ^ postmask
	if np.count_nonzero(hitmask & premask):
		print("Que? Some ranges disappeared?")
else:
	# no --domain flag, do a plain sweep
	hitmask = do_sweep(maskrange)

al = u.adt.build_addr_lookup()
for start, stop in describe_mask(hitmask, maskrange):
	# bit ugly but it makes addrlookup do all the heavy lifting for us
	al.add(range(start, stop), "hit")

print("Hits:")
for zone, value in al.items():
	if ((zone.start - 1) // granule + 1) * granule >= zone.stop:
		continue
	if not any([v[0] == "hit" for v in value]):
		continue

	labels = set([v[0] for v in value if v[0] != "hit"])
	print(f"\t{zone.start:9x} - {zone.stop:9x} | {' '.join(labels)}")



from m1n1.utils import Register8, RegMap
from enum import IntEnum

class R_PWR(Register8):
	APWDN_EN  = 7
	BSNS_PWDN = 6
	S_RST     = 1
	SPWDN     = 0

class R_GEC(Register8):
	EDGE     = 4
	ANA_GAIN = 1, 0

class E_FS(IntEnum):
	FS_8_12    = 0b000
	FS_16_24   = 0b001
	FS_32_48   = 0b010
	FS_64_96   = 0b011
	FS_128_192 = 0b100
	FS_48_72   = 0b101

class R_DAC(Register8):
	HV   = 7
	MUTE = 6
	HPF  = 5
	LPM  = 4
	FS   = 1, 0, E_FS

class R_SAI1(Register8):
	DAC_POL    = 7
	BCLK_POL   = 6
	TDM_BCLKS  = 5, 3
	FSYNC_MODE = 2
	SDATA_FMT  = 1
	SAI_MODE   = 0

class R_SAI2(Register8):
	DATA_WIDTH = 7
	AUTO_SLOT  = 4
	TDM_SLOT   = 3, 0

class R_STATUS(Register8):
	UVLO_VREG = 6
	LIM_EG    = 5
	CLIP      = 4
	AMP_OC    = 3
	OTF       = 2
	OTW       = 1
	BAT_WARN  = 0

class SSM3515Regs(RegMap):
	PWR      = 0x00, R_PWR
	GEC      = 0x01, R_GEC
	DAC      = 0x02, R_DAC
	DAC_VOL  = 0x03, Register8
	SAI1     = 0x04, R_SAI1
	SAI2     = 0x05, R_SAI2
	VBAT_OUT = 0x06, Register8
	LIM1     = 0x07, Register8
	LIM2     = 0x08, Register8
	LIM3     = 0x09, Register8
	STATUS   = 0x0a, Register8
	FAULT    = 0x0b, Register8



# SPDX-License-Identifier: MIT

import struct

from m1n1.hw.dart import DART
from m1n1.hw.ane import ANERegs, ANEDARTRegs, ANETaskManager


class ANE:

	PAGE_SIZE = 0x4000
	TILE_SIZE = 0x4000

	def __init__(self, u):
		self.u = u
		self.p = u.proxy

		self.name = "ane"
		self.p.pmgr_adt_clocks_enable(f'/arm-io/{self.name}')
		self.p.pmgr_adt_clocks_enable(f'/arm-io/dart-{self.name}')

		self.base_addr = u.adt[f'arm-io/{self.name}'].get_reg(0)[0]
		self.regs = ANERegs(self.u, self.base_addr)
		self.apply_static_tunables()

		ps_map = {
			"ane":  0x023b70c000,
			"ane0": 0x028e08c000,
			"ane1": 0x028e684000,
			"ane2": 0x228e08c000,
			"ane3": 0x228e684000,
		}
		self.ps_base_addr = ps_map[self.name]

		# we need a slight patch to dart
		self.dart = DART.from_adt(u, path=f'/arm-io/dart-{self.name}',
					instance=0, iova_range=(0x4000, 0xe0000000))
		self.dart.initialize()
		dart_regs = []
		for prop in range(3):
			dart_addr = self.u.adt[f'/arm-io/dart-{self.name}'].get_reg(prop)[0]
			dart_regs.append(ANEDARTRegs(self.u, dart_addr))
		self.dart_regs = dart_regs

		# hack to initialize base ttbr
		phys = self.u.memalign(self.PAGE_SIZE, self.PAGE_SIZE)
		self.dart.iomap_at(0, 0x0, phys, self.PAGE_SIZE)
		self.ttbr0_addr = self.dart.regs.TTBR[0, 0].val
		self.dart_regs[1].TTBR[0, 0].val = self.ttbr0_addr  # DMA fails w/o
		self.dart_regs[2].TTBR[0, 0].val = self.ttbr0_addr  # DMA fails w/o

		self.allocator = ANEAllocator(self)
		self.fw = ANEFirmware(self)
		self.tm = ANETaskManager(self)

	def apply_static_tunables(self):  # this cost me a solid week
		static_tunables_map = [
			(0x0, 0x10), (0x38, 0x50020), (0x3c, 0xa0030),
			(0x400, 0x40010001), (0x600, 0x1ffffff),
			(0x738, 0x200020), (0x798, 0x100030),
			(0x7f8, 0x100000a), (0x900, 0x101), (0x410, 0x1100),
			(0x420, 0x1100), (0x430, 0x1100)]
		for (offset, value) in static_tunables_map:
			self.p.write32(self.base_addr + offset, value)

	def power_up(self):
		self.p.pmgr_adt_clocks_enable(f'/arm-io/{self.name}')
		self.p.pmgr_adt_clocks_enable(f'/arm-io/dart-{self.name}')
		self.power_down()
		for offset in range(0x0, 0x30+0x8, 0x8):
			self.p.write32(self.ps_base_addr + offset, 0xf)
		self.tm.reset()

	def power_down(self):
		for offset in reversed(range(0x0, 0x30+0x8, 0x8)):
			self.p.write32(self.ps_base_addr + offset, 0x300)

	def ioread(self, iova, size):
		return self.dart.ioread(0, iova & 0xffffffff, size)

	def iowrite(self, iova, buf):
		self.dart.iowrite(0, iova & 0xffffffff, buf)

	def round_up(self, x, y): return ((x+(y-1)) & (-y))


class ANEBuffer:
	def __init__(self, mapid, phys, iova, size):
		self.mapid = mapid
		self.phys = phys
		self.iova = iova
		self.size = size


class ANEAllocator:
	def __init__(self, ane):
		self.ane = ane
		self.mapid = 0
		self.map = {}

	def alloc_size(self, size):
		size = self.ane.round_up(size, self.ane.PAGE_SIZE)
		phys = self.ane.u.memalign(self.ane.PAGE_SIZE, size)
		self.ane.p.memset32(phys, 0, size)
		iova = self.ane.dart.iomap(0, phys, size)

		buf = ANEBuffer(self.mapid, phys, iova, size)
		self.map[self.mapid] = buf
		print("mapid %d: mapped phys 0x%x to iova 0x%06x for data w/ size 0x%06x"
			  % (buf.mapid, buf.phys, buf.iova, buf.size))
		self.mapid += 1
		return buf.iova

	def alloc_data(self, data):
		iova = self.alloc_size(len(data))
		self.ane.iowrite(iova, data)
		return iova

	def dump_map(self):
		for mapid in self.map:
			buf = self.map[mapid]
			print('mapid %d: phys 0x%x, iova 0x%x, size 0x%x'
				  % (buf.mapid, buf.phys, buf.iova, buf.size))


class ANEEngineReq:
	def __init__(self, anec):
		self.anec = anec
		self.td_size = 0
		self.td_count = 0
		self.fifo_iova = 0
		self.nid = 0
		self.qid = 0
		self.bar = [0x0] * 0x20


class ANEFirmware:

	FIFO_NID = 0x40
	FIFO_COUNT = 0x20
	FIFO_WIDTH = 0x400  # nextpow2(0x274)

	def __init__(self, ane):
		self.ane = ane

	def setup(self, anec):
		req = ANEEngineReq(anec)
		req.td_size = anec.td_size
		req.td_count = anec.td_count

		# setup immutable bar
		tsk_buf = anec.data[anec.tsk_start:anec.tsk_start+anec.tsk_size]
		req.bar[0] = self.ane.allocator.alloc_data(tsk_buf)
		krn_start = anec.tsk_start + self.ane.round_up(anec.tsk_size, 0x10)
		krn_buf = anec.data[krn_start:krn_start+anec.krn_size]
		req.bar[1] = self.ane.allocator.alloc_data(krn_buf)

		# setup mutable bar
		for bdx in range(0x20):
			if ((anec.tiles[bdx]) and (bdx >= 3)):
				size = anec.tiles[bdx] * self.ane.TILE_SIZE
				req.bar[bdx] = self.ane.allocator.alloc_size(size)

		self.make_fifo(req)
		return req

	def make_fifo(self, req):
		anec = req.anec
		pool_size = self.ane.round_up(self.FIFO_WIDTH * 2, self.ane.TILE_SIZE)
		fifo_iova = self.ane.allocator.alloc_size(pool_size)

		td_buf = anec.data[anec.tsk_start:anec.tsk_start+anec.td_size]
		fifo_head = self.set_nid(td_buf, self.FIFO_NID)
		fifo_tail = self.set_nid(td_buf, self.FIFO_NID + self.FIFO_COUNT)
		self.ane.iowrite(fifo_iova, fifo_head)
		self.ane.iowrite(fifo_iova + self.FIFO_WIDTH, fifo_tail)

		req.fifo_iova = fifo_iova
		req.nid = self.FIFO_NID
		req.qid = 4  # just the default queue

	def set_nid(self, td_buf, nid):
		hdr0 = struct.unpack('<L', td_buf[:4])[0]
		hdr0 = (hdr0 & 0xf00ffff) | ((nid & 0xff) << 16)
		return struct.pack('<L', hdr0) + td_buf[4:]

	def send_src(self, req, src_buf, idx):
		iova = req.bar[4 + req.anec.dst_count + idx]
		size = req.anec.src_sizes[idx]
		if (len(src_buf) < size):
			src_buf += b''*(size - len(src_buf))
		self.ane.iowrite(iova, src_buf[:size])

	def read_dst(self, req, idx):
		iova = req.bar[4 + idx]
		size = req.anec.dst_sizes[idx]
		return self.ane.ioread(iova, size)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.constructutils import Ver
from m1n1.utils import *

Ver.set_version(u)

from m1n1.shell import run_shell

from m1n1.fw.agx import Agx

p.pmgr_adt_clocks_enable("/arm-io/gfx-asc")
p.pmgr_adt_clocks_enable("/arm-io/sgx")

agx = Agx(u)
agx.verbose = 10

#agx.uat.dump(0)

agx.build_initdata()

run_shell(globals(), msg="Have fun!")



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1 import asm

p.smp_start_secondaries()

tfreq = u.mrs(CNTFRQ_EL0)

TEST_CPUS = [1, 4]

CLUSTER_PSTATE = 0x20020
CLUSTER_STATUS = 0x20050

chip_id = u.adt["/chosen"].chip_id

if chip_id in (0x8103, 0x6000, 0x6001, 0x6002):
    CREG = [
        0x210e00000,
        0x211e00000,
    ]

    MAX_PSTATE = [5, 15]

elif chip_id in (0x8121, 0x6020, 0x6021, 0x6022):
    CREG = [
        0x210e00000,
        0x211e00000,
    ]

    if u.adt["/chosen"].target_type == "J416c":
        MAX_PSTATE = [7, 19]
    else:
        MAX_PSTATE = [7, 17]

code = u.malloc(0x1000)

util = asm.ARMAsm(f"""
bench:
    mrs x1, CNTPCT_EL0
1:
    sub x0, x0, #1
    cbnz x0, 1b

    mrs x2, CNTPCT_EL0
    sub x0, x2, x1
    ret

signal_and_write:
    sev
    mrs x2, CNTPCT_EL0
    add x2, x2, #0x800
1:
    mrs x3, CNTPCT_EL0
    sub x4, x3, x2
    cbnz x4, 1b
    str x1, [x0]
    mov x0, x3
    ret

timelog:
    mrs x2, s3_1_c15_c0_0 /* SYS_IMP_APL_PMCR0 */
    orr x2, x2, #1
    msr s3_1_c15_c0_0, x2
    mov x2, #0xffffffffffffffff
    msr s3_1_c15_c1_0, x2
    isb
    wfe
1:
    mrs x2, CNTPCT_EL0
    mrs x3, s3_2_c15_c0_0
    isb
    stp x2, x3, [x0], #16
    mov x4, #0x40
2:
    sub x4, x4, #1
    cbnz x4, 2b
    sub x1, x1, #1
    cbnz x1, 1b
    
    ret
""", code)
iface.writemem(code, util.data)
p.dc_cvau(code, len(util.data))
p.ic_ivau(code, len(util.data))

def bench_cpu(idx, loops=10000000):
    if idx == 0:
        elapsed = p.call(util.bench, loops) / tfreq
    else:
        elapsed = p.smp_call_sync(idx, util.bench, loops) / tfreq
    if elapsed == 0:
        return 0
    mhz = (loops / elapsed) / 1000000
    return mhz

def set_pstate(cluster, pstate):
    p.mask64(CREG[cluster] + CLUSTER_PSTATE, 0x1f01f, (1<<25) | pstate)

print()

LOG_ITERS = 10000
logbuf = u.malloc(LOG_ITERS * 16)

def bench_latency(cluster, cpu, from_pstate, to_pstate, verbose=False):
    set_pstate(cluster, from_pstate)
    bench_cpu(cpu)

    p.smp_call(cpu, util.timelog, logbuf, LOG_ITERS)
    psreg = (p.read64(CREG[cluster] + CLUSTER_PSTATE) & ~0x1f01f) | (1<<25) | to_pstate
    tval = p.call(util.signal_and_write, CREG[cluster] + CLUSTER_PSTATE, psreg)
    p.smp_wait(cpu)
    
    logdata = iface.readmem(logbuf, LOG_ITERS * 16)
    lts, lcyc = None, None
    
    log = []
    for i in range(LOG_ITERS):
        ts, cyc = struct.unpack("<QQ", logdata  [i*16:i*16+16])
        log.append((ts, cyc))

    off = 256
    
    ts_0, cyc_0 = log[off]
    ts_e, cyc_e = log[-1]
    f_init = None
    f_end = None
    lts, lcyc = ts_0, cyc_0

    inc = to_pstate > from_pstate

    blip = 0
    cnt = dts_sum = 0
    for i in range(off, len(log)):
        ts, cyc = log[i]
        dts = ts - lts
        dcyc = cyc - lcyc

        cnt += 1
        dts_sum += dts

        blip = max(blip, dts)

        if f_init is None and ts > tval:
            tidx = i
            f_init = (lcyc - cyc_0) / (lts - ts_0) * tfreq / 1000000
            dts_init = dts_sum / cnt
        if f_end is None and ts > (tval + ts_e) / 2:
            f_end = (cyc_e - cyc) / (ts_e - ts) * tfreq / 1000000
            cnt = dts_sum = 0
    
        #if lts is not None:
            #print(f"{i}: {ts}: {cyc} ({ts-lts}: {cyc-lcyc})")
        #else:
            #print(f"{i}: {ts}: {cyc}")
        lts, lcyc = ts, cyc

    dts_end = dts_sum / cnt

    window = 32

    if verbose:
        print(f"Triggered at {tval}")

    thresh = 2/ (1/f_init + 1/f_end)

    for i in range(tidx, LOG_ITERS - window - 1):
        ts0, cyc0 = log[i - window]
        ts1, cyc1 = log[i + window]
        f = (cyc1 - cyc0) / (ts1 - ts0) * tfreq / 1000000
        if inc and (f > thresh) or ((not inc) and f < thresh):
            tts = log[i][0]
            tidx = i
            if verbose:
                print(f"Frequency transition at #{i} {tts}")
            break

    if verbose:
        print(f"Initial frequency: {f_init:.2f}")
        print(f"Final frequency: {f_end:.2f}")
        print(f"Threshold: {thresh:.2f}")

        for i in range(max(window, tidx - 10 * window), tidx + 10 * window):
            ts0, cyc0 = log[i - window]
            ts1, cyc1 = log[i + window]
            lts, lcyc = log[i - 1]
            ts, cyc = log[i]
            f = (cyc1 - cyc0) / (ts1 - ts0) * tfreq / 1000000
            print(f"{i}: {ts}: {cyc} ({ts-lts}: {cyc-lcyc}): {f:.2f}")

    blip -= min(dts_init, dts_end)

    return (tts - tval) / tfreq * 1000000000, blip / tfreq * 1000000000

for cluster, creg in enumerate(CREG):
    cpu = TEST_CPUS[cluster]

    freqs = []

    print(f"#### Cluster {cluster} ####")
    print(" P-States:")
    print("  ", end="")
    for pstate in range(MAX_PSTATE[cluster] + 1):
        set_pstate(cluster, pstate)
        freq = int(round(bench_cpu(cpu)))
        freqs.append(freq)
        print(f"{pstate}:{freq}MHz", end=" ")
    print()
    print()
    
    print(" To-> |", end="")
    for to_pstate in range(1, MAX_PSTATE[cluster] + 1):
        print(f" {freqs[to_pstate]:7d} |", end="")
    print()
    print(" From |", end="")
    for to_pstate in range(1, MAX_PSTATE[cluster] + 1):
        print(f"---------+", end="")
    print()
    
    maxblip = 0
    
    for from_pstate in range(1, MAX_PSTATE[cluster] + 1):
        print(f" {freqs[from_pstate]:4d} |", end="")
        for to_pstate in range(1, MAX_PSTATE[cluster] + 1):
            if from_pstate == to_pstate:
                print(f" ******* |", end="")
                continue
            lat, blip = bench_latency(cluster, cpu, from_pstate, to_pstate)
            print(f" {lat:7.0f} |", end="")
            maxblip = max(maxblip, blip)
        print()
    
    print()
    print(f"Maximum execution latency spike: {maxblip:.0f} ns")
    print()

print()

#bench_latency(1, TEST_CPUS[1], 15, 14, True)





# SPDX-License-Identifier: MIT

import struct
from construct import *

from m1n1.hv import TraceMode
from m1n1.proxyutils import RegMonitor
from m1n1.utils import *

from m1n1.trace import ADTDevTracer
from m1n1.trace.asc import ASCTracer, ASCRegs, EP, EPState, msg, msg_log, DIR
from m1n1.trace.dart import DARTTracer
from m1n1.trace.gpio import GPIOTracer
from m1n1.trace.spi import SPITracer

DARTTracer = DARTTracer._reloadcls()
ASCTracer = ASCTracer._reloadcls()
SPITracer = SPITracer._reloadcls()

# SPI HID transport tracer for 2021 macbook models

kbd_node = None
for node in hv.adt.walk_tree():
    try:
        if node.compatible[0] == "spi-1,spimc":
            for c in node:
                try:
                    if c.compatible[0] == "hid-transport,k1":
                        kbd_node = c
                        break
                except AttributeError:
                    continue
    except AttributeError:
        continue
    if kbd_node is not None:
        break


class Z2Tracer(SPITracer):
    def start(self):
        super().start()
        self.txbuffer = []
        self.rxbuffer = []
        self.want_bytes = 0
        self.state = Z2Tracer.preboot
    def w_TXDATA(self, data):
        self.txbuffer.append(data.value)
        self.check_msg_finished()
    def r_RXDATA(self, data):
        self.rxbuffer.append(data.value)
        self.check_msg_finished()
    def check_msg_finished(self):
        if min(len(self.txbuffer), len(self.rxbuffer)) < self.want_bytes:
            return
        self.state(self)
    def bad_state(self):
        pass
    def error(self):
        self.log(f"RXBUF {' '.join(hex(x) for x in self.rxbuffer)}")
        self.log(f"TXBUF {' '.join(hex(x) for x in self.txbuffer)}")
        self.log(f"state: {self.state}")
        self.log("Tracer desynchronized, shutting down")
        self.state = Z2Tracer.bad_state
    def consume_bytes(self, n):
        self.txbuffer = self.txbuffer[n:]
        self.rxbuffer = self.rxbuffer[n:]
    def preboot(self):
        if self.txbuffer[0] == 0:
            self.want_bytes = 4
            self.state = Z2Tracer.init_zeros
        elif self.txbuffer[0] == 0x1e:
            self.want_bytes = 16
            self.state = Z2Tracer.processing_init_data
        else:
            self.error()
    def init_zeros(self):
        self.log("sent 4 zeroes")
        self.consume_bytes(4)
        self.state = Z2Tracer.preboot
    def processing_init_data(self):
        self.log("Sent init data")
        self.want_bytes = 2
        self.consume_bytes(16)
        self.state = Z2Tracer.main_hbpp
    def main_hbpp(self):
        if self.txbuffer[0] == 0x1a and self.txbuffer[1] == 0xa1:
            self.log("Sent int ack")
            self.consume_bytes(2)
        elif self.txbuffer[0] == 0x18 and self.txbuffer[1] == 0xe1:
            self.log("Sent nop")
            self.consume_bytes(2)
        elif self.txbuffer[0] == 0x1f and self.txbuffer[1] == 0x01:
            self.log("Sent request cal")
            self.consume_bytes(2)
        elif self.txbuffer[0] == 0x30 and self.txbuffer[1] == 0x01:
            self.state = Z2Tracer.send_blob_cmd
            self.want_bytes = 10
        elif self.txbuffer[0] == 0x1e and self.txbuffer[1] == 0x33:
            self.state = Z2Tracer.send_rmw_cmd
            self.want_bytes = 16
        elif self.txbuffer[0] == 0xee and self.txbuffer[1] == 0x00:
            self.state = Z2Tracer.main_z2
            self.want_bytes = 16
        else:
            self.error()
    def send_blob_cmd(self):
        length = (self.txbuffer[2] << 8 | self.txbuffer[3]) * 4
        self.consume_bytes(10)
        self.want_bytes = length + 4
        self.log(f"Sending blob of length {length}")
        self.state = Z2Tracer.send_blob_tail
    def send_blob_tail(self):
        self.log("Finished sendind blob")
        self.consume_bytes(self.want_bytes)
        self.want_bytes = 2
        self.state = Z2Tracer.main_hbpp
    def send_rmw_cmd(self):
        self.log('Sent RMW command')
        self.want_bytes = 2
        self.consume_bytes(16)
        self.state = Z2Tracer.main_hbpp
    def main_z2(self):
        if self.txbuffer[0] == 0xee:
            self.log("sent wake cmd")
            self.consume_bytes(16)
        elif self.txbuffer[0] == 0xe2:
            self.log("sent get device info cmd")
            self.consume_bytes(16)
            self.state = Z2Tracer.read_device_info_reply
        elif self.txbuffer[0] == 0xeb:
            length = (self.rxbuffer[1] | (self.rxbuffer[2] << 8)) + 5
            length = (length + 3) & (-4)
            self.consume_bytes(16)
            self.want_bytes = length
            self.state = Z2Tracer.read_interrupt_data
        elif self.txbuffer[0] == 0xe3:
            self.log(f"got report info for {self.txbuffer[1]}, len is {self.rxbuffer[3]}")
            self.consume_bytes(16)
        elif self.txbuffer[0] == 0xe7:
            self.want_bytes = self.txbuffer[3] + 5
            self.consume_bytes(16)
            self.state = Z2Tracer.reading_report_long
        elif self.txbuffer[0] == 0xe6:
            self.consume_bytes(16)
            self.state = Z2Tracer.read_report_reply
        else:
            self.error()
    def reading_report_long(self):
        self.log(f"got report {' '.join(hex(x) for x in self.rxbuffer)}")
        self.consume_bytes(self.want_bytes)
        self.want_bytes = 16
        self.state = Z2Tracer.main_z2
    def read_interrupt_data(self):
        data = self.rxbuffer[5:]
        tstamp2 = data[4] | (data[5] << 8) | (data[6] << 16)
        tx = [f"TS1 {hex(data[1])} TS2 {tstamp2} UNK1: {mxformat(data[7:16])} UNK2: {mxformat(data[17:24])}"]
        if len(data) >= 16:
            ntouch = data[16]
            for i in range(ntouch):
                ptr = 24 + 30 * i
                finger = data[ptr]
                state = data[ptr + 1]
                x = data[ptr + 4] | (data[ptr + 5] << 8) 
                y = data[ptr + 6] | (data[ptr + 7] << 8)
                wj = data[ptr + 12] | (data[ptr + 13] << 8)
                wn = data[ptr + 14] | (data[ptr + 15] << 8)
                dg = data[ptr + 16] | (data[ptr + 17] << 8)
                prs = data[ptr + 18] | (data[ptr + 19] << 8)
                tx.append(f"F: {hex(finger)} S: {hex(state)} X: {x} Y: {y} MAJ: {wj} MIN: {wn} ANG: {dg} PRS: {prs} UNK1: {mxformat(data[ptr + 2:ptr+4])} UNK2: {mxformat(data[ptr + 8:ptr+12])} UNK3: {mxformat(data[ptr + 20:ptr+30])}")
            self.log(';'.join(tx))
        else:
            self.log(f"??? {mxformat(data)}")
        self.consume_bytes(self.want_bytes)
        self.want_bytes = 16
        self.state = Z2Tracer.main_z2
    def read_device_info_reply(self):
        self.log(f"got device info {' '.join(hex(x) for x in self.rxbuffer[:16])}")
        self.consume_bytes(16)
        self.state = Z2Tracer.main_z2
    def read_report_reply(self):
        self.log(f"got report {' '.join(hex(x) for x in self.rxbuffer[:16])}")
        self.consume_bytes(16)
        self.state = Z2Tracer.main_z2

def mxformat(ls):
    return ''.join(xformat(x) for x in ls)
def xformat(x):
    x = hex(x)[2:]
    if len(x) == 1:
        x = '0' + x
    return x



# trace interrupts
aic_phandle = getattr(hv.adt["/arm-io/aic"], "AAPL,phandle")
spi_node = kbd_node._parent

#if getattr(spi_node, "interrupt-parent") == aic_phandle:
#    for irq in getattr(spi_node, "interrupts"):
#        hv.trace_irq(node.name, irq, 1, hv.IRQTRACE_IRQ)
#for irq in hv.adt['/arm-io/gpio'].interrupts:
#    hv.trace_irq('/arm-io/gpio', irq, 1, hv.IRQTRACE_IRQ)

spi_tracer = Z2Tracer(hv, "/arm-io/" + spi_node.name)
spi_tracer.start()

spi_pins_nub = {
    0x0: "clock32khz",
}

#gpio_tracer_nub = GPIOTracer(hv, "/arm-io/nub-gpio", spi_pins_nub, verbose=0)
#gpio_tracer_nub.start()

spi_pins = {
    0x6d: "enable_cs",
    0x8b: "reset"
}

#gpio_tracer = GPIOTracer(hv, "/arm-io/gpio", spi_pins, verbose=0)
#gpio_tracer.start()
